import unittest
import os

import numpy as np

import hazenlib.utils as hazen_tools
from hazenlib.utils import dcmread
from tests import TEST_DATA_DIR


class ShapeSetUp(unittest.TestCase):
    SMALL_CIRCLE_PHANTOM_FILE = str(
        TEST_DATA_DIR
        / "ghosting"
        / "PE_COL_PHANTOM_BOTTOM_RIGHT"
        / "PE_COL_PHANTOM_BOTTOM_RIGHT.IMA"
    )
    small_circle_x, small_circle_y, small_circle_r = 1, 1, 22.579364776611328

    LARGE_CIRCLE_PHANTOM_FILE = str(
        TEST_DATA_DIR / "uniformity" / "axial_oil.IMA"
    )
    large_circle_x, large_circle_y, large_circle_r = (
        128,
        123,
        97.84805297851562,
    )

    SAG_RECTANGLE_PHANTOM_FILE = str(TEST_DATA_DIR / "uniformity" / "sag.dcm")
    rectangle_size = (204, 177.0)  # (Vertical, Horizontal)
    rectangle_angle = -90
    rectangle_centre = (130.500015, 134.499985)

    COR_RECTANGLE_PHANTOM_FILE = str(TEST_DATA_DIR / "uniformity" / "cor.dcm")
    cor_rectangle_size = (206.047546, 194.684875)
    cor_rectangle_angle = -89.1756591796875
    cor_rectangle_centre = (128.43869, 136.219971)

    COR2_RECTANGLE_PHANTOM_FILE = str(
        TEST_DATA_DIR / "uniformity" / "cor2.dcm"
    )
    cor2_rectangle_size = (194.4591522216797, 201.483292)
    cor2_rectangle_angle = -1.576546
    cor2_rectangle_centre = (127.261551, 130.001953)


# @pytest.mark.skip
class TestShapeDetector(ShapeSetUp):
    def setUp(self) -> None:
        pass

    def test_large_circle(self):
        arr = dcmread(self.LARGE_CIRCLE_PHANTOM_FILE).pixel_array
        shape_detector = hazen_tools.ShapeDetector(arr=arr)
        x, y, r = shape_detector.get_shape("circle")
        assert int(x), int(y) == (self.large_circle_x, self.large_circle_y)
        assert round(r) == round(self.large_circle_r)

    def test_small_circle(self):
        arr = dcmread(self.SMALL_CIRCLE_PHANTOM_FILE).pixel_array
        shape_detector = hazen_tools.ShapeDetector(arr=arr)
        x, y, r = shape_detector.get_shape("circle")
        assert int(x), int(y) == (self.small_circle_x, self.small_circle_y)
        assert round(r) == round(self.small_circle_r)

    def test_sag_rectangle(self):
        arr = dcmread(self.SAG_RECTANGLE_PHANTOM_FILE).pixel_array
        shape_detector = hazen_tools.ShapeDetector(arr=arr)
        centre, size, angle = shape_detector.get_shape("rectangle")
        np.testing.assert_allclose(centre, self.rectangle_centre, rtol=1e-02)
        np.testing.assert_allclose(size, self.rectangle_size, rtol=1e-02)
        np.testing.assert_allclose(angle, self.rectangle_angle, rtol=1e-02)

    def test_cor_rectangle(self):
        arr = dcmread(self.COR_RECTANGLE_PHANTOM_FILE).pixel_array
        shape_detector = hazen_tools.ShapeDetector(arr=arr)
        centre, size, angle = shape_detector.get_shape("rectangle")
        np.testing.assert_allclose(
            centre, self.cor_rectangle_centre, rtol=1e-02
        )
        np.testing.assert_allclose(size, self.cor_rectangle_size, rtol=1e-02)
        np.testing.assert_allclose(angle, self.cor_rectangle_angle, rtol=1e-02)

    def test_cor2_rectangle(self):
        arr = dcmread(self.COR2_RECTANGLE_PHANTOM_FILE).pixel_array
        shape_detector = hazen_tools.ShapeDetector(arr=arr)
        centre, size, angle = shape_detector.get_shape("rectangle")
        np.testing.assert_allclose(
            centre, self.cor2_rectangle_centre, rtol=1e-02
        )
        np.testing.assert_allclose(size, self.cor2_rectangle_size, rtol=1e-02)
        np.testing.assert_allclose(
            angle, self.cor2_rectangle_angle, rtol=1e-02
        )


class Test_is_Dicom_file(unittest.TestCase):
    def setUp(self) -> None:
        data_folder = str(TEST_DATA_DIR / "tools")
        self.true_dicom_path = os.path.join(data_folder, "dicom_yes.dcm")
        self.false_dicom_path = os.path.join(data_folder, "dicom_no.jfif")

    def test_is_dicom(self):
        result = hazen_tools.is_dicom_file(self.true_dicom_path)
        self.assertTrue(result)

        result = hazen_tools.is_dicom_file(self.false_dicom_path)
        self.assertFalse(result)

    def test_is_dicom_yes(self):
        result = hazen_tools.is_dicom_file(self.true_dicom_path)
        self.assertTrue(result)

    def test_is_dicom_no(self):
        result = hazen_tools.is_dicom_file(self.false_dicom_path)
        self.assertFalse(result)


class TestUtils(unittest.TestCase):
    def setUp(self):
        TEST_DICOM = str(
            TEST_DATA_DIR / "toshiba" / "TOSHIBA_TM_MR_DCM_V3_0.dcm"
        )
        TEST_DICOM = dcmread(TEST_DICOM)
        print(TEST_DICOM.Columns * TEST_DICOM.PixelSpacing[0])
        self.test_dicoms = {
            "philips": {
                "file": str(
                    TEST_DATA_DIR
                    / "resolution"
                    / "philips"
                    / "IM-0004-0002.dcm"
                ),
                "MANUFACTURER": "philips",
                "ROWS": 512,
                "COLUMNS": 512,
                "TR_CHECK": 500,
                "BW": 205.0,
                "ENHANCED": False,
                "PIX_ARRAY": 1,
                "SLICE_THICKNESS": 5,
                "PIX_SIZE": [0.48828125, 0.48828125],
                "AVERAGE": 1,
            },
            "siemens": {
                "file": str(
                    TEST_DATA_DIR
                    / "resolution"
                    / "resolution_site01"
                    / "256_sag.IMA"
                ),
                "MANUFACTURER": "siemens",
                "ROWS": 256,
                "COLUMNS": 256,
                "TR_CHECK": 500,
                "BW": 130.0,
                "ENHANCED": False,
                "PIX_ARRAY": 1,
                "SLICE_THICKNESS": 5,
                "PIX_SIZE": [0.9765625, 0.9765625],
                "AVERAGE": 1,
            },
            "toshiba": {
                "file": str(
                    TEST_DATA_DIR / "toshiba" / "TOSHIBA_TM_MR_DCM_V3_0.dcm"
                ),
                "MANUFACTURER": "toshiba",
                "ROWS": 256,
                "COLUMNS": 256,
                "TR_CHECK": 45.0,
                "BW": 244.0,
                "ENHANCED": False,
                "PIX_ARRAY": 1,
                "SLICE_THICKNESS": 6,
                "PIX_SIZE": [1.0, 1.0],
                "AVERAGE": 1,
            },
            "ge": {
                "file": str(TEST_DATA_DIR / "ge" / "ge_eFilm.dcm"),
                "MANUFACTURER": "ge",
                "ROWS": 256,
                "COLUMNS": 256,
                "TR_CHECK": 1000.0,
                "BW": 156.25,
                "ENHANCED": False,
                "PIX_ARRAY": 1,
                "SLICE_THICKNESS": 5,
                "PIX_SIZE": [0.625, 0.625],
                "AVERAGE": 1,
            },
        }

    def test_is_enhanced(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                enhanced = hazen_tools.is_enhanced_dicom(dcm)
                assert enhanced == self.test_dicoms[manufacturer]["ENHANCED"]

    def test_get_manufacturer(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                assert (
                    hazen_tools.get_manufacturer(dcm)
                    == self.test_dicoms[manufacturer]["MANUFACTURER"]
                )

    def get_average(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                avg = hazen_tools.get_average(dcm)
                assert avg == self.test_dicoms[manufacturer]["AVERAGE"]

    def test_get_bandwidth(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                bw = hazen_tools.get_bandwidth(dcm)
                assert bw == self.test_dicoms[manufacturer]["BW"]

    def test_get_num_of_frames(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                pix_arr = hazen_tools.get_num_of_frames(dcm)
                assert pix_arr == self.test_dicoms[manufacturer]["PIX_ARRAY"]

    def test_get_slice_thickness(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                slice_thick = hazen_tools.get_slice_thickness(dcm)
                assert (
                    slice_thick
                    == self.test_dicoms[manufacturer]["SLICE_THICKNESS"]
                )

    def get_pixel_size(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                pix_size = hazen_tools.get_pixel_size(dcm)
                pix_size = list(pix_size)
                self.assertEqual(
                    pix_size, self.test_dicoms[manufacturer]["PIX_SIZE"]
                )

    def test_get_TR(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                TR = hazen_tools.get_TR(dcm)
                assert TR == self.test_dicoms[manufacturer]["TR_CHECK"]

    def test_get_rows(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                rows = hazen_tools.get_rows(dcm)
                assert rows == self.test_dicoms[manufacturer]["ROWS"]

    def test_get_columns(self):
        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                columns = hazen_tools.get_columns(dcm)
                assert columns == self.test_dicoms[manufacturer]["COLUMNS"]

    def test_fov(self):
        self.test_dicoms = {
            "philips": {
                "file": str(
                    TEST_DATA_DIR
                    / "resolution"
                    / "philips"
                    / "IM-0004-0002.dcm"
                ),
                "MANUFACTURER": "philips",
                "FOV": 250.0,
            },
            "siemens": {
                "file": str(
                    TEST_DATA_DIR
                    / "resolution"
                    / "resolution_site01"
                    / "256_sag.IMA"
                ),
                "MANUFACTURER": "siemens",
                "FOV": 250.0,
            },
            "toshiba": {
                "file": str(
                    TEST_DATA_DIR / "toshiba" / "TOSHIBA_TM_MR_DCM_V3_0.dcm"
                ),
                "MANUFACTURER": "toshiba",
                "FOV": 256.0,
            },
        }

        for manufacturer in self.test_dicoms.keys():
            with dcmread(self.test_dicoms[manufacturer]["file"]) as dcm:
                # first test function
                fov = hazen_tools.get_field_of_view(dcm)
                print(fov)
                assert fov == self.test_dicoms[manufacturer]["FOV"]

    # def test_get_image_orientation(self):
    #     # TODO add unit test for image orientation

    def test_rescale_to_byte(self):
        test_array = np.array([[1, 2], [3, 4]])
        TEST_OUT = np.array([[63, 127], [191, 255]])
        test_array = hazen_tools.rescale_to_byte(test_array)
        test_array = test_array.tolist()
        TEST_OUT = TEST_OUT.tolist()
        self.assertListEqual(test_array, TEST_OUT)


class TestNewDicom(unittest.TestCase):
    """Tests for new_dicom function that creates DICOM copies with updated pixel data."""

    def setUp(self):
        self.test_file = str(
            TEST_DATA_DIR / "toshiba" / "TOSHIBA_TM_MR_DCM_V3_0.dcm"
        )
        self.original_dcm = dcmread(self.test_file)
        self.original_pixels = self.original_dcm.pixel_array.copy()

    def test_creates_deep_copy(self):
        """Test that new_dicom creates an independent copy, not a reference."""
        modified_pixels = self.original_pixels // 2

        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, modified_pixels, instance_number=1
        )

        # Original should be unchanged
        np.testing.assert_array_equal(
            self.original_dcm.pixel_array, self.original_pixels
        )
        # New DICOM should have modified data
        np.testing.assert_array_equal(new_dcm.pixel_array, modified_pixels)
        # Should be different objects
        self.assertIsNot(new_dcm, self.original_dcm)

    def test_sets_instance_number(self):
        """Test that instance_number is correctly set in the new DICOM."""
        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, self.original_pixels, instance_number=42
        )

        self.assertEqual(new_dcm.InstanceNumber, 42)

    def test_none_instance_number(self):
        """Test handling of None instance_number."""
        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, self.original_pixels, instance_number=None
        )

        self.assertIsNone(new_dcm.InstanceNumber)

    def test_preserves_critical_metadata(self):
        """Test that PatientName, Study IDs and other metadata are preserved."""
        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, self.original_pixels, instance_number=1
        )

        self.assertEqual(new_dcm.PatientName, self.original_dcm.PatientName)
        self.assertEqual(
            new_dcm.StudyInstanceUID, self.original_dcm.StudyInstanceUID
        )
        self.assertEqual(
            new_dcm.SeriesInstanceUID, self.original_dcm.SeriesInstanceUID
        )
        self.assertEqual(new_dcm.SOPClassUID, self.original_dcm.SOPClassUID)
        self.assertEqual(new_dcm.Rows, self.original_dcm.Rows)
        self.assertEqual(new_dcm.Columns, self.original_dcm.Columns)

    def test_photometric_interpretation_preserved(self):
        """Test that PhotometricInterpretation tag is preserved."""
        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, self.original_pixels, instance_number=1
        )

        self.assertEqual(
            new_dcm.PhotometricInterpretation,
            self.original_dcm.PhotometricInterpretation,
        )

    def test_bits_stored_preserved(self):
        """Test that BitsStored tag is preserved."""
        new_dcm = hazen_tools.new_dicom(
            self.original_dcm, self.original_pixels, instance_number=1
        )

        self.assertEqual(new_dcm.BitsStored, self.original_dcm.BitsStored)


class TestSplitDicom(unittest.TestCase):
    """Tests for split_dicom function that handles multiframe DICOMs."""

    def setUp(self):
        # Use non-enhanced DICOM for base tests
        self.non_enhanced_file = str(
            TEST_DATA_DIR / "toshiba" / "TOSHIBA_TM_MR_DCM_V3_0.dcm"
        )
        self.non_enhanced_dcm = dcmread(self.non_enhanced_file)

    def test_non_enhanced_returns_single_element_list(self):
        """Test that non-enhanced DICOM returns list with single original element."""
        result = hazen_tools.split_dicom(self.non_enhanced_dcm)

        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        # Should be the same object (not a copy) for efficiency
        self.assertIs(result[0], self.non_enhanced_dcm)

    def test_non_enhanced_preserves_pixel_array(self):
        """Test that pixel array is unchanged for non-enhanced split."""
        result = hazen_tools.split_dicom(self.non_enhanced_dcm)

        np.testing.assert_array_equal(
            result[0].pixel_array, self.non_enhanced_dcm.pixel_array
        )

    def test_non_enhanced_preserves_instance_number(self):
        """Test that original InstanceNumber is preserved for non-enhanced."""
        original_instance_num = self.non_enhanced_dcm.InstanceNumber
        result = hazen_tools.split_dicom(self.non_enhanced_dcm)

        self.assertEqual(result[0].InstanceNumber, original_instance_num)

    def test_enhanced_splitting(self):
        # Requires enhanced multiframe test data
        enhanced_file = str(
            TEST_DATA_DIR / "acr" / "Siemens_Sola_1.5T_T1/" / "1.dcm",
        )
        enhanced_dcm = dcmread(enhanced_file)

        # Skip if not actually enhanced
        if not hazen_tools.is_enhanced_dicom(enhanced_dcm):
            self.skipTest("Test requires enhanced multiframe DICOM")

        result = hazen_tools.split_dicom(enhanced_dcm)

        # Should split into NumberOfFrames elements
        self.assertEqual(len(result), enhanced_dcm.NumberOfFrames)

        # Each should have InstanceNumber set to frame index (1-based)
        for i, frame_dcm in enumerate(result, 1):
            self.assertEqual(frame_dcm.InstanceNumber, i)
            # Each should be 2D (original is 3D)
            self.assertEqual(len(frame_dcm.pixel_array.shape), 2)


if __name__ == "__main__":
    unittest.main()



import logging
import pathlib
import unittest

from hazenlib.tasks.acr_spatial_resolution import ACRSpatialResolution
from hazenlib.utils import get_dicom_files

from tests import TEST_DATA_DIR, TEST_REPORT_DIR

# noqa: S101

logger = logging.getLogger(__name__)


class TestACRSpatialResolutionSiemens(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    EXPECTED_RESOLVED_ROWS = [3, 2, -1, 2, 3, 2]
    EXPECTED_UNRESOLVED_ARRAY = [1.0]
    EXPECTED_UL_RESOLUTION = 0.9
    EXPECTED_LR_RESOLUTION = 0.9
    EXPECTED_RESOLUTION = 0.9

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)

        self.acr_spatial_resolution_task = ACRSpatialResolution(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

        self.dcm = self.acr_spatial_resolution_task.ACR_obj.slice_stack[0]
        self.detected_rows = (
            self.acr_spatial_resolution_task.get_spatially_resolved_rows(
                self.dcm
            )
        )
        self.resolved_arrays = (
            self.acr_spatial_resolution_task.get_resolved_arrays(
                self.detected_rows
            )
        )
        self.ul_resolution, self.lr_resolution, self.best_resolution = (
            self.acr_spatial_resolution_task.get_best_resolution(
                self.resolved_arrays
            )
        )

    def test_get_best_resolution(self):
        assert self.EXPECTED_UL_RESOLUTION == self.ul_resolution
        assert self.EXPECTED_LR_RESOLUTION == self.lr_resolution
        assert self.EXPECTED_RESOLUTION == self.best_resolution

    def test_unresolved_array(self):
        unresolved_resolutions = [
            k for k, v in self.resolved_arrays.items() if not v["resolved"]
        ]
        assert unresolved_resolutions == self.EXPECTED_UNRESOLVED_ARRAY

    def test_get_spatially_resolved_rows(self):
        assert self.EXPECTED_RESOLVED_ROWS == self.detected_rows


class TestACRSpatialResolutionSiemensSolaFit(TestACRSpatialResolutionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFit")
    EXPECTED_RESOLVED_ROWS = [1, 3, 1, 2, 1, 3]
    EXPECTED_UNRESOLVED_ARRAY = []
    EXPECTED_UL_RESOLUTION = 0.9
    EXPECTED_LR_RESOLUTION = 0.9
    EXPECTED_RESOLUTION = 0.9


class TestACRSpatialResolutionPhilipsAchieva(TestACRSpatialResolutionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchieva")
    EXPECTED_RESOLVED_ROWS = [1, 1, 1, 1, -1, -1]
    EXPECTED_UNRESOLVED_ARRAY = [0.9]
    EXPECTED_UL_RESOLUTION = 1.0
    EXPECTED_LR_RESOLUTION = 1.0
    EXPECTED_RESOLUTION = 1.0


class TestACRSpatialResolutionGE(TestACRSpatialResolutionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    EXPECTED_RESOLVED_ROWS = [1, 1, 1, 1, 1, 1]
    EXPECTED_UNRESOLVED_ARRAY = []
    EXPECTED_UL_RESOLUTION = 0.9
    EXPECTED_LR_RESOLUTION = 0.9
    EXPECTED_RESOLUTION = 0.9



"""Tests for hazenlib types and data structures."""

# ruff: noqa: PT009 PT027

from __future__ import annotations

# Type Checking
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

# Python imports
import json
import unittest
from dataclasses import FrozenInstanceError
from unittest.mock import MagicMock, patch

# Module imports
import numpy as np

# Local imports
from hazenlib.types import (
    Measurement,
    Metadata,
    Result,
    TaskMetadata,
)


class TestJsonSerializableMixin(unittest.TestCase):
    """Tests for JSON serialization base class."""

    def test_to_dict_converts_nested_objects(self) -> None:
        """Verify to_dict recursively converts dataclass instances."""
        result = Result(task="TestTask", desc="test", files=["file.dcm"])
        result.add_measurement(
            Measurement(name="SNR", value=42.0, visibility="final"),
        )

        d = result.to_dict()

        self.assertIsInstance(d, dict)
        self.assertEqual(d["task"], "TestTask")
        self.assertIsInstance(d["measurements"], list)
        self.assertEqual(d["measurements"][0]["name"], "SNR")

    def test_to_dict_handles_numpy_arrays(self) -> None:
        """Verify numpy arrays are converted to lists in to_dict."""
        arr = np.array([1.0, 2.0, 3.0])
        result = Result(task="Test", desc="test", files=["f.dcm"])
        result.add_measurement(
            Measurement(name="SNR", value=arr, visibility="final"),
        )

        d = result.to_dict()

        self.assertIsInstance(d["measurements"][0]["value"], np.ndarray)
        self.assertTrue(
            np.all(d["measurements"][0]["value"] == [1.0, 2.0, 3.0]),
        )

    def test_to_json_returns_valid_json(self) -> None:
        """Verify to_json produces parseable JSON string."""
        result = Result(task="JSONTest", desc="json test", files=[])
        result.add_measurement(
            Measurement(name="SNR", value=1.0, visibility="final"),
        )

        json_str = result.to_json()

        # Should be valid JSON
        parsed = json.loads(json_str)
        self.assertEqual(parsed["task"], "JSONTest")
        self.assertEqual(parsed["measurements"][0]["value"], 1.0)

    def test_to_dict_handles_empty_collections(self) -> None:
        """Verify empty measurements and metadata are handled gracefully."""
        result = Result(task="Empty", desc="empty", files=[])
        # No measurements added

        d = result.to_dict()

        self.assertEqual(d["measurements"], [])
        self.assertEqual(d["report_images"], [])


class TestMeasurement(unittest.TestCase):
    """Tests for the Measurement dataclass."""

    def test_measurement_creation(self) -> None:
        """Verify Measurement can be created with all fields."""
        m = Measurement(
            name="SNR",
            value=42.5,
            type="measured",
            subtype="ratio",
            description="Signal to noise ratio",
            unit="dB",
            visibility="final",
        )

        self.assertEqual(m.name, "SNR")
        self.assertEqual(m.value, 42.5)
        self.assertEqual(m.type, "measured")
        self.assertEqual(m.visibility, "final")

    def test_measurement_defaults(self) -> None:
        """Verify Measurement has sensible defaults for optional fields."""
        m = Measurement(name="SNR", value=1.0)

        self.assertEqual(m.type, "measured")
        self.assertEqual(m.subtype, "")
        self.assertEqual(m.description, "")
        self.assertEqual(m.unit, "")
        self.assertEqual(m.visibility, "final")  # Assuming default is final

    def test_measurement_is_frozen(self) -> None:
        """Verify Measurement is immutable (frozen dataclass)."""
        m = Measurement(name="SNR", value=1.0)

        with self.assertRaises(FrozenInstanceError):
            m.name = "NewName"  # type: ignore[misc]

        with self.assertRaises(FrozenInstanceError):
            m.value = 2.0  # type: ignore[misc]

    def test_measurement_accepts_numpy_values(self) -> None:
        """Verify Measurement can store numpy scalar types."""
        np_value = np.float64(42.5)
        m = Measurement(
            name="SNR",
            value=np_value,
            visibility="final",
        )

        self.assertIsInstance(m.value, np.float64)
        self.assertEqual(float(m.value), 42.5)


class TestResult(unittest.TestCase):
    """Tests for the Result dataclass and its methods."""

    def test_result_initialization(self) -> None:
        """Verify Result initializes with correct default values."""
        result = Result(
            task="TestTask",
            desc="test desc",
            files=["a.dcm", "b.dcm"],
        )

        self.assertEqual(result.task, "TestTask")
        self.assertEqual(result.desc, "test desc")
        self.assertEqual(result.files, ["a.dcm", "b.dcm"])
        self.assertEqual(result.measurements, ())
        self.assertEqual(set(result.report_images), set())
        self.assertIsInstance(result.metadata, Metadata)

    def test_add_measurement(self) -> None:
        """Verify add_measurement appends to internal list."""
        result = Result(task="Test", desc="test", files=["f.dcm"])
        m1 = Measurement(name="SNR", value=1.0, visibility="final")
        m2 = Measurement(name="Ghosting", value=2.0, visibility="intermediate")

        result.add_measurement(m1)
        result.add_measurement(m2)

        self.assertEqual(len(result.measurements), 2)
        self.assertEqual(result.measurements[0].name, "SNR")
        self.assertEqual(result.measurements[1].name, "Ghosting")

    def test_add_report_image(self) -> None:
        """Verify add_report_image adds paths to internal list."""
        result = Result(task="Test", desc="test", files=["input.dcm"])

        result.add_report_image("output1.png")
        result.add_report_image("output2.png")

        self.assertEqual(
            set(result.report_images),
            {"output1.png", "output2.png"},
        )

    def test_get_measurement_filtering(self) -> None:
        """Verify get_measurement filters correctly by attributes."""
        result = Result(task="Test", desc="test", files=["f.dcm"])
        result.add_measurement(
            Measurement(name="SNR", value=10.0, visibility="final"),
        )
        result.add_measurement(
            Measurement(
                name="SNR",
                type="raw",
                value=20.0,
                visibility="intermediate",
            ),
        )
        result.add_measurement(
            Measurement(name="Ghosting", value=5.0, visibility="final"),
        )

        # Filter by name
        snr_measurements = result.get_measurement(name="SNR")
        self.assertEqual(len(snr_measurements), 2)

        # Filter by combined criteria
        specific = result.get_measurement(
            name="SNR",
            measurement_type="raw",
        )
        self.assertEqual(len(specific), 1)
        self.assertEqual(specific[0].value, 20.0)

    def test_get_measurement_empty_result(self) -> None:
        """Verify get_measurement returns empty list when no matches."""
        result = Result(task="SNR", desc="test", files=[])

        matches = result.get_measurement(name="nonexistent")
        self.assertEqual(matches, [])

    def test_filtered_creates_copy(self) -> None:
        """Verify filtered() returns new Result without modifying original."""
        result = Result(task="Test", desc="test", files=["f.dcm"])
        result.add_measurement(
            Measurement(name="SNR", value=1.0, visibility="final"),
        )
        result.add_measurement(
            Measurement(
                name="Ghosting",
                value=2.0,
                visibility="intermediate",
            ),
        )

        filtered = result.filtered(level="final")

        # Original should be unchanged
        self.assertEqual(len(result.measurements), 2)
        # Filtered should have only one
        self.assertEqual(len(filtered.measurements), 1)
        self.assertEqual(filtered.measurements[0].name, "SNR")
        # Metadata copied
        self.assertEqual(filtered.task, "Test")

    def test_filtered_all_returns_self(self) -> None:
        """Verify filtered(level='all') returns self for efficiency."""
        result = Result(task="Test", desc="test", files=["f.dcm"])
        result.add_measurement(
            Measurement(name="SNR", value=1.0, visibility="final"),
        )

        filtered = result.filtered(level="all")

        # Should be the same object (identity check)
        self.assertIs(filtered, result)

    def test_filtered_invalid_level_raises_error(self) -> None:
        """Verify filtered raises ValueError for invalid level."""
        result = Result(task="Test", desc="test", files=["f.dcm"])

        with self.assertRaises(ValueError) as context:
            result.filtered(level="invalid_level")

        self.assertIn("Invalid measurement visibility", str(context.exception))


class TestMetadata(unittest.TestCase):
    """Tests for the Metadata dataclass."""

    def test_metadata_defaults(self) -> None:
        """Verify Metadata initializes with None/empty defaults."""
        m = Metadata()

        self.assertIsNone(m.institution_name)
        self.assertIsNone(m.manufacturer)
        self.assertIsNone(m.model)
        self.assertIsNone(m.date)
        self.assertIsNone(m.series_id)
        self.assertIsNone(m.study_id)
        self.assertIsNotNone(m.version)

    def test_metadata_with_values(self) -> None:
        """Verify Metadata accepts construction with values."""
        m = Metadata(
            institution_name="Test Hospital",
            manufacturer="GE",
            model="Artist",
        )

        self.assertEqual(m.institution_name, "Test Hospital")
        self.assertEqual(m.manufacturer, "GE")
        self.assertEqual(m.model, "Artist")


class TestTaskMetadata(unittest.TestCase):
    """Tests for TaskMetadata configuration class."""

    def test_task_metadata_defaults(self) -> None:
        """Verify TaskMetadata has correct defaults."""
        tm = TaskMetadata(module_name="Test", class_name="TestClass")

        self.assertEqual(tm.single_image, False)
        self.assertIsNone(tm.phantom)
        self.assertIsNone(tm.requires_args)


class TestMetadataDicomExtraction(unittest.TestCase):
    """Tests for Metadata __post_init__ DICOM extraction."""

    def setUp(self) -> None:
        """Set up realistic GE scanner DICOM attributes."""
        self.ge_dicom_attrs = {
            "InstitutionName": "Anon",  # From your DICOM dump
            "Manufacturer": "GE MEDICAL SYSTEMS",
            "ManufacturerModelName": "Signa HDxt",
            "StudyDate": "20200903",
            "SeriesInstanceUID": (
                "1.2.840.113619.2.322.2807.4256007.22970.1599113147.28"
            ),
            "StudyInstanceUID": (
                "1.2.840.113619.6.322.29258601288877239892320825850091988423"
            ),
            "AcquisitionNumber": "1",
        }

    def _create_mock_dicom(self, **overrides: dict[str, Any]) -> MagicMock:
        """Create a mock pydicom Dataset with GE-like attributes."""
        mock = MagicMock()
        attrs = {**self.ge_dicom_attrs, **overrides}
        for key, value in attrs.items():
            setattr(mock, key, value)
        # Ensure hasattr works correctly
        mock.__contains__ = lambda _, key: key in attrs
        return mock

    @patch("hazenlib.types.pydicom.dcmread")
    def test_extracts_ge_scanner_metadata(
        self,
        mock_dcmread: Callable,
    ) -> None:
        """Verify all metadata fields extracted from GE DICOM."""
        mock_dcmread.return_value = self._create_mock_dicom()

        m = Metadata(files=["tests/data/acr/GE/0.dcm"])

        self.assertEqual(m.institution_name, "Anon")
        self.assertEqual(m.manufacturer, "GE MEDICAL SYSTEMS")
        self.assertEqual(m.model, "Signa HDxt")
        self.assertEqual(m.date, "20200903")
        self.assertEqual(
            m.series_id,
            "1.2.840.113619.2.322.2807.4256007.22970.1599113147.28",
        )
        self.assertEqual(
            m.study_id,
            "1.2.840.113619.6.322.29258601288877239892320825850091988423",
        )
        self.assertEqual(m.acquisition_number, "1")

    @patch("hazenlib.types.pydicom.dcmread")
    def test_skips_extraction_if_fields_populated(
        self,
        mock_dcmread: Callable,
    ) -> None:
        """Verify existing values are preserved, not overwritten."""
        mock_dcmread.return_value = self._create_mock_dicom(
            InstitutionName="Different Hospital",
        )

        m = Metadata(
            files=["test.dcm"],
            institution_name="Keep This Value",
            manufacturer="Siemens",  # Also set differently
        )

        # Should keep provided values, not extract from DICOM
        self.assertEqual(m.institution_name, "Keep This Value")
        self.assertEqual(m.manufacturer, "Siemens")
        # But unset fields should still extract
        self.assertEqual(m.model, "Signa HDxt")

    @patch("hazenlib.types.pydicom.dcmread")
    def test_handles_missing_optional_attributes(
        self,
        mock_dcmread: Callable,
    ) -> None:
        """Verify extraction works when optional DICOM tags are missing."""
        # Create DICOM without InstitutionName (anonymized/removed)
        mock = self._create_mock_dicom()
        delattr(mock, "InstitutionName")

        mock_dcmread.return_value = mock

        m = Metadata(files=["anon.dcm"])

        # Should still extract other fields
        self.assertEqual(m.manufacturer, "GE MEDICAL SYSTEMS")
        self.assertIsNone(m.institution_name)  # Not present in DICOM

    @patch("hazenlib.types.pydicom.dcmread")
    def test_warns_on_multiple_different_values(
        self,
        mock_dcmread: Callable,
    ) -> None:
        """Verify warning logged when files have different institutions."""
        mock_dcmread.side_effect = [
            self._create_mock_dicom(InstitutionName="Hospital A"),
            self._create_mock_dicom(InstitutionName="Hospital B"),
            self._create_mock_dicom(InstitutionName="Hospital C"),
        ]

        with self.assertLogs("hazenlib.types", level="WARNING") as log_context:
            m = Metadata(files=["1.dcm", "2.dcm", "3.dcm"])

        # Should warn about multiple values
        self.assertTrue(
            any(
                "institution_name" in msg and "Multiple" in msg
                for msg in log_context.output
            ),
        )
        # Should still pick one value (arbitrary from set)
        self.assertIn(
            m.institution_name,
            ["Hospital A", "Hospital B", "Hospital C"],
        )

    @patch("hazenlib.types.pydicom.dcmread")
    def test_graceful_on_dicom_read_error(
        self,
        mock_dcmread: Callable,
    ) -> None:
        """Verify Metadata creation succeeds even if files can"t be read."""
        mock_dcmread.side_effect = Exception("Permission denied")

        # Should not raise, just log warning
        with self.assertLogs("hazenlib.types", level="WARNING"):
            m = Metadata(files=["corrupt.dcm"])

        # Fields remain None (or version defaults)
        self.assertIsNone(m.institution_name)
        self.assertIsNone(m.manufacturer)

    @patch("hazenlib.types.pydicom.dcmread")
    def test_no_extraction_without_files(self, mock_dcmread: Callable) -> None:
        """Verify dcmread is never called when files is None."""
        m = Metadata()  # No files

        mock_dcmread.assert_not_called()
        self.assertIsNone(m.institution_name)



"""Orchestration Module for performing multiple tasks."""

# ruff: noqa: T201

from __future__ import annotations

import datetime
import hashlib

# Type Checking
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    # Python imports
    from collections.abc import Sequence

    # Local imports
    from hazenlib.HazenTask import HazenTask

# Python imports
import importlib
import logging
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import TypeVar

# Module imports
import yaml
from docx import Document
from docx.shared import Inches
from packaging import version as packaging_version
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from pydicom import dcmread

# Local imports
from hazenlib._version import __version__
from hazenlib.ACRObject import ACRObject
from hazenlib.exceptions import (
    UnknownAcquisitionTypeError,
    UnknownTaskNameError,
)
from hazenlib.types import Measurement, PhantomType, Result, TaskMetadata
from hazenlib.utils import get_dicom_files, wait_on_parallel_results

logger = logging.getLogger(__name__)


def init_task(
    selected_task: str,
    files: list[str],
    report: bool = False,
    report_dir: str | None = None,
    **kwargs,
) -> HazenTask:
    """Initialise object of the correct HazenTask class.

    Args:
        selected_task (string): name of task script/module to load
        files (list): list of filepaths to DICOM images
        report (bool): whether to generate report images
        report_dir (string): path to folder to save report images to
        kwargs: any other key word arguments

    Returns:
        an object of the specified HazenTask class

    """
    try:
        meta = TASK_REGISTRY[selected_task]
    except KeyError as err:
        msg = f"Unknown task: {selected_task}"
        logger.exception(
            "%s. Supported tasks are:\n%s",
            msg,
            "\n\t".join(TASK_REGISTRY),
        )
        raise ValueError(msg) from err

    # Import module
    task_module = importlib.import_module(f"hazenlib.tasks.{meta.module_name}")

    # Get explicit class
    try:
        task_class = getattr(task_module, meta.class_name)
    except AttributeError as err:
        msg = f"Module {meta.module_name} has no class '{meta.class_name}'"
        raise ImportError(msg) from err

    return task_class(
        input_data=files,
        report=report,
        report_dir=report_dir,
        **kwargs,
    )


class AcquisitionType(Enum):
    """Supported Acquisition Types."""

    ACR_T1 = "ACR T1"
    ACR_T2 = "ACR T2"
    ACR_SL = "ACR Sagittal Localizer"

    @classmethod
    def from_string(cls, value: str) -> AcquisitionType:
        """Create AcquisitionType from string with fuzzy matching.

        Handles British/American spelling variations (Localiser vs Localizer)
        and case insensitivity.

        Args:
            value: String representation of acquisition type

        Returns:
            AcquisitionType enum value

        Raises:
            UnknownAcquisitionTypeError: If string cannot be matched to
                a known type

        Example:
            >>> AcquisitionType.from_string("t1")
            AcquisitionType.ACR_T1
            >>> AcquisitionType.from_string("sagittal localiser")
            AcquisitionType.ACR_SL

        """
        normalized = value.lower().replace("localiser", "localizer")
        mapping = {
            "t1": cls.ACR_T1,
            "t2": cls.ACR_T2,
            "sagittal localizer": cls.ACR_SL,
            "sagittal": cls.ACR_SL,
            "localizer": cls.ACR_SL,
        }
        if normalized in mapping:
            return mapping[normalized]
        raise UnknownAcquisitionTypeError(value)


@dataclass(frozen=True)
class ProtocolStep:
    """Single protocol step representing the execution of a single task.

    Attributes:
        task_name: Name of the task as registered in TASK_REGISTRY
        acquisition_type: Type of acquisition this step applies to

    """

    task_name: str
    acquisition_type: AcquisitionType

    def __post_init__(self) -> None:
        """Validate that task name exists in the global registry."""
        if self.task_name not in TASK_REGISTRY:
            available = ", ".join(sorted(TASK_REGISTRY.keys()))
            raise UnknownTaskNameError(self.task_name, available)


@dataclass
class Protocol:
    """Orchestrator for collections of tasks.

    Attributes:
        name: Human-readable identifier for this protocol
        steps: Ordered collection of protocol steps to execute

    """

    name: str
    steps: tuple[ProtocolStep, ...] = field(default_factory=tuple)

    @classmethod
    def from_config(cls, config_path: Path) -> Protocol:
        """Load protocol from configuration file."""
        msg = "The class method 'from_config' has not been implemented yet."
        raise NotImplementedError(msg)


class ProtocolResult(Result):
    """Class for the protocol result."""

    def __post_init__(self, _load_metadata: bool) -> None:
        """Initialise the results list."""
        super().__post_init__(_load_metadata)

        # Set initial result to contain
        # protocol information.
        self._results: list[Result] = [
            Result(
                self.task,
                self.desc,
                self.files,
            ),
        ]

    @property
    def measurements(self) -> tuple[Measurement, ...]:
        """Tuple of initial result measurements."""
        return self.results[0].measurements

    def add_measurement(self, measurement: Measurement) -> None:
        """Add a measurement to the initial result."""
        self.results[0].add_measurement(measurement)

    @property
    def report_images(self) -> tuple[str, ...]:
        """Tuple of initial report image locations."""
        return self.results[0].report_images

    def add_report_image(self, image_path: str | Sequence[str]) -> None:
        """Add a report image to the initial result."""
        return self.results[0].add_report_image(image_path)

    @property
    def results(self) -> tuple[Result, ...]:
        """Return an immutable results list."""
        return tuple(self._results)

    def add_result(self, result: Result) -> None:
        """Add a result to the list."""
        self._results.append(result)

    def add_metadata_to_doc(self, doc: Document) -> Document:
        """Add metadata properties to a document."""
        doc.core_properties.author = "Hazen"
        doc.core_properties.created = datetime.datetime.now(tz=datetime.UTC)
        doc.core_properties.comments = (
            "Initial draft of the document automatically generated by"
            f" Hazen (version {self.metadata.version})"
        )
        return doc

    def to_docx(
        self,
        template_path: Path | str | None = None,
        level: str = "all",
    ) -> Document:
        """Generate Word document from aggregated results."""
        if self.results is None:
            msg = "Results cannot be empty"
            logger.error(
                "%s, please make sure the protocols run method"
                " has been called.",
                msg,
            )
            raise ValueError(msg)

        if template_path is not None:
            template_path = Path(template_path)
            if not template_path.exists():
                msg = f"Template file not found: {template_path}"
                logger.error(msg)
                raise FileNotFoundError(msg)
            doc = Document(template_path)
        else:
            doc = Document()

            # Header with protocol info
            doc.add_heading(f"QA Report: {self.task}", 0)

        # Metadata
        doc = self.add_metadata_to_doc(doc)
        # Section per step
        for _result in self.results:
            # Skip the metadata result for the current protocol
            if _result.task == self.task:
                continue
            doc = add_report_table_to_doc(doc, _result, level)

        return doc


def add_report_table_to_doc(
    doc: Document,
    _result: Result,
    level: str,
) -> Document:
    """Add a report table to the word document."""
    doc.add_heading(_result.task, level=1)

    # Filter out specific results ('all', 'final', 'intermediate', etc.)
    result = _result.filtered(level)

    # internal Measurement property -> word heading
    text_mapping = {
        "name": "Name",
        "type": "Type",
        "subtype": "Subtype",
        "description": "Description",
        "value": "Value",
        "unit": "Unit",
    }

    # Results table
    table = doc.add_table(rows=1, cols=len(text_mapping))
    table.style = "Light Grid Accent 1"
    hdr_cells = table.rows[0].cells
    for idx, text in enumerate(text_mapping.values()):
        hdr_cells[idx].text = text

    for m in result.measurements:
        row_cells = table.add_row().cells
        for idx, key in enumerate(text_mapping.keys()):
            value = getattr(m, key)
            row_cells[idx].text = str(value) if value is not None else "-"

    # Embed report images if generated
    if level != "final":  # Ignore for final reports.
        for img_path in result.report_images:
            if Path(img_path).exists():
                doc.add_picture(img_path, width=Inches(5.0))
    return doc


T = TypeVar("T")


class ACRLargePhantomProtocol(Protocol):
    """Protocol for ACR Large Phantom."""

    def __init__(self, dirs: list[str | Path], **kwargs: T.kwargs) -> None:
        """Set up the protocol for the ACR Large Phantom."""
        self.name = "ACR Large Phantom"
        self.steps = (
            # Geometric Accuracy.
            ProtocolStep("acr_geometric_accuracy", AcquisitionType.ACR_T1),
            ProtocolStep(
                "acr_sagittal_geometric_accuracy",
                AcquisitionType.ACR_SL,
            ),
            # High Contrast Object Detection.
            ProtocolStep("acr_spatial_resolution", AcquisitionType.ACR_T1),
            ProtocolStep("acr_spatial_resolution", AcquisitionType.ACR_T2),
            # Slice Thickness Accuracy.
            ProtocolStep("acr_slice_thickness", AcquisitionType.ACR_T1),
            ProtocolStep("acr_slice_thickness", AcquisitionType.ACR_T2),
            # Slice Position Accuracy.
            ProtocolStep("acr_slice_position", AcquisitionType.ACR_T1),
            ProtocolStep("acr_slice_position", AcquisitionType.ACR_T2),
            # Image Intensity Uniformity.
            ProtocolStep("acr_uniformity", AcquisitionType.ACR_T1),
            ProtocolStep("acr_uniformity", AcquisitionType.ACR_T2),
            # Percent Signal Ghosting.
            ProtocolStep("acr_ghosting", AcquisitionType.ACR_T1),
            # Low Contrast Object Detectability.
            ProtocolStep(
                "acr_low_contrast_object_detectability",
                AcquisitionType.ACR_T1,
            ),
            ProtocolStep(
                "acr_low_contrast_object_detectability",
                AcquisitionType.ACR_T2,
            ),
            # SNR
            ProtocolStep("acr_snr", AcquisitionType.ACR_T1),
            ProtocolStep("acr_snr", AcquisitionType.ACR_T2),
        )

        kwargs.setdefault("report", False)
        kwargs.setdefault("report_dir", None)
        self.kwargs = kwargs

        required_acquisition_types = {s.acquisition_type for s in self.steps}
        if len(dirs) != (num_aq := len(required_acquisition_types)):
            msg = f"Incorrect number of directories - should be {num_aq}"
            logger.exception("%s but got %i", msg, len(dirs))
            raise ValueError(msg)

        files_list = (get_dicom_files(d) for d in dirs)
        self.file_groups = {}
        for files in files_list:
            acr_obj = ACRObject([dcmread(files[0], stop_before_pixels=False)])
            acquisition_type = AcquisitionType.from_string(
                acr_obj.acquisition_type(strict=True),
            )
            self.file_groups[acquisition_type] = files

        if len(self.file_groups.keys()) != len(required_acquisition_types):
            acquisition_types = set(self.file_groups.keys())
            msg = (
                f"Missing sequences - only {acquisition_types}"
                " were found meaning"
                f" {required_acquisition_types - acquisition_types}"
                " were not present."
            )
            logger.exception(
                "%s The following directories were passed: %s",
                msg,
                dirs,
            )
            raise ValueError(msg)

    def run(self, *, debug: bool = False) -> ProtocolResult:
        """Run the Protocol for each of the steps."""
        results = ProtocolResult(
            self.name,
            ", ".join(
                f"{s.task_name} {s.acquisition_type}" for s in self.steps
            ),
            list(self.file_groups.values()),
        )

        arg_list = [
            (step, self.file_groups, self.kwargs) for step in self.steps
        ]
        parallel_results = wait_on_parallel_results(
            _execute_step,
            arg_list,
            debug=debug,
        )
        for r in parallel_results:
            results.add_result(r)

        return results


def _execute_step(
    step: ProtocolStep,
    file_groups: dict,
    kwargs: T.kwargs,
) -> Result:
    """Encapsulate the work for a single step."""
    task = init_task(
        step.task_name,
        file_groups[step.acquisition_type],
        **kwargs,
    )
    return task.run()


@dataclass(frozen=True)
class JobTaskConfig:
    """Configuration for a specific task within a job.

    Attributes:
        task : Must match the TASK_REGISTRY or the PROTOCOL_REGISTRY key.
        folders : Path to the folders containing the DICOMs.
        overrides : Task specific overrides.

    """

    task: str
    folders: list[Path]
    overrides: dict[str, Any] = field(default_factory=dict)
    _is_protocol: bool = field(init=False, compare=False, repr=False)

    def __post_init__(self) -> None:
        """Initialise the job task configuration."""
        if self.task in TASK_REGISTRY:
            object.__setattr__(self, "_is_protocol", False)
        elif self.task in PROTOCOL_REGISTRY:
            object.__setattr__(self, "_is_protocol", True)
            # Protocol-specific validation
            if self.task == "acr_all" and len(self.folders) != 3:
                msg = (
                    "acr_all expects 3 directories (T1, T2, Sagittal),"
                    f" got {len(self.folders)}"
                )
                raise ValueError(msg)
        else:
            available_tasks = ", ".join(TASK_REGISTRY.keys())
            available_protocols = ",".join(PROTOCOL_REGISTRY.keys())
            raise UnknownTaskNameError(
                self.task,
                f"{available_protocols}, {available_tasks}",
            )

        for folder in self.folders:
            if not Path(folder).exists():
                msg = f"Folder not found: {folder}"
                raise FileNotFoundError(msg)

    @property
    def is_protocol(self) -> bool:
        """Return True if this job references a protocol."""
        return self._is_protocol


@dataclass
class BatchConfig:
    """Configuration for the batch configuration file."""

    version: str
    hazen_version_constraint: str | None
    description: str
    jobs: list[JobTaskConfig]
    output: str
    report_docx: str | None = None
    report_template: str | None = None
    defaults: dict[str, Any] | None = None
    levels: list[str] | tuple[str] = ("final", "all")

    _file: str | Path | None = field(default=None)
    _dry_run: bool = False

    _CURRENT_BATCHCONFIG_VERSION: str = "1.0"

    def to_yaml(self, output: Path | str) -> None:
        """Save the batch configuration object to a yaml file.

        Args:
            output: Path to write the YAML configuration file.

        """

        def resolve_path_as_posix(path: str | Path | None) -> str | None:
            if path is None:
                return None
            return Path(path).absolute().as_posix()

        output = Path(output)
        output.parent.mkdir(parents=True, exist_ok=True)

        # Build the data dictionary in the expected order
        data: dict[str, Any] = {
            "version": self.version,
        }

        data["hazen_version_constraint"] = self.hazen_version_constraint

        data["description"] = self.description

        data["output"] = resolve_path_as_posix(self.output)

        data["levels"] = list(self.levels)

        data["report_docx"] = resolve_path_as_posix(self.report_docx)

        data["report_template"] = resolve_path_as_posix(self.report_template)

        data["defaults"] = {} if self.defaults is None else self.defaults

        data["defaults"]["report"] = self._check_and_set_default_report(
            data["levels"],
            report=data["defaults"].get("report", False),
        )

        # Build jobs list
        jobs_data: list[dict[str, Any]] = []
        for job in self.jobs:
            job_dict: dict[str, Any] = {
                "task": job.task,
                "folders": [resolve_path_as_posix(f) for f in job.folders],
            }
            if job.overrides:
                job_dict["overrides"] = {}
                for k, v in job.overrides.items():
                    if isinstance(v, Path):
                        job_dict["overrides"][k] = resolve_path_as_posix(v)
                    else:
                        job_dict["overrides"][k] = v
            jobs_data.append(job_dict)

        data["jobs"] = jobs_data

        with output.open("w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)

    def __post_init__(self) -> None:
        """Log the batch config initial parameters."""
        logger.debug(
            "Performing batch config job from file: %s",
            self._file,
        )

    def run(
        self,
        *,
        dry_run: bool | None = None,
        debug: bool = False,
    ) -> ProtocolResult:
        """Run the batch config tasks."""
        dry_run = self._dry_run if dry_run is None else dry_run

        ############
        # Protocol #
        ############

        protocol_jobs = [j for j in self.jobs if j.is_protocol]
        protocol_arg_list = []
        for job in protocol_jobs:
            kwargs = dict(self.defaults or {})
            kwargs.update(job.overrides)

            folders = job.folders
            protocol_arg_list.append([job.task, folders, kwargs])

        #########
        # Tasks #
        #########

        task_jobs = [j for j in self.jobs if not j.is_protocol]
        task_arg_list = []
        for job in task_jobs:
            kwargs = dict(self.defaults or {})
            kwargs.update(job.overrides)

            files = get_dicom_files(job.folders[0])
            task_arg_list.append([job.task, files, kwargs])

        result_files = [] if self._file is None else [self._file.as_posix()]
        results = ProtocolResult(
            task="Batch Configuration Job",
            desc=self.description,
            files=result_files,
        )

        ###########
        # Dry run #
        ###########

        if dry_run:
            print(f"Configuration valid. Would execute {len(self.jobs)} jobs:")

            # Protocols
            for i, (protocol, folders, kwargs) in enumerate(protocol_arg_list):
                print(
                    f"{i}. Protocol: {protocol}\n"
                    f"\tFolders:     {[f.as_posix() for f in folders]}\n"
                    f"\tParameters:  {kwargs or '(none)'}",
                )

            # Tasks
            for i, (task, files, kwargs) in enumerate(task_arg_list):
                print(
                    f"{i + len(protocol_arg_list)}. Task: {task}\n"
                    f"\tFiles:       {len(files)} DICOM(s)"
                    f" from {Path(files[0]).parent}\n"
                    f"\tParameters:  {kwargs or '(none)'}",
                )
            print("-" * 60 + "\nDry run complete. No Measurements performed.")
            return results

        #######
        # Run #
        #######

        # Protocols
        for job, args in zip(protocol_jobs, protocol_arg_list, strict=True):
            _, dirs, kwargs = args
            protocol = PROTOCOL_REGISTRY[job.task](dirs, **kwargs)
            protocol_results = protocol.run()
            for r in protocol_results.results:
                if r.task == protocol_results.task:
                    continue
                results.add_result(r)

        # Tasks
        parallel_results = wait_on_parallel_results(
            _execute_task,
            task_arg_list,
            debug=debug,
        )

        for r in parallel_results:
            results.add_result(r)

        return results

    @classmethod
    def _migrate_config(
        cls,
        data: dict[str, Any],
        schema_version: str,
    ) -> dict[str, Any]:
        """Migrate the schema version."""
        msg = "Configuration schema migration is not currently supported."
        raise NotImplementedError(msg)

    @classmethod
    def _validate_schema_version(
        cls,
        schema_version: str,
        data: dict[str, Any],
    ) -> dict[str, Any]:
        current_schema = cls._CURRENT_BATCHCONFIG_VERSION
        if packaging_version.parse(schema_version) > packaging_version.parse(
            current_schema,
        ):
            msg = (
                f"Config file schema version {schema_version} is newer than "
                f"supported version {current_schema}"
            )
            logger.error(
                "%s. Please upgrade hazen to use this configuration file.",
                msg,
            )
            raise ValueError(msg)
        if schema_version != current_schema:
            logger.warning(
                "Config file uses schema version %s"
                ", current is %s."
                " Attempting backward-compatible load."
                " Consider updating your config file to the latest schema.",
                schema_version,
                current_schema,
            )
            data = cls._migrate_config(data, schema_version)
        return data

    @staticmethod
    def _validate_hazen_version(constraint_str: str) -> None:
        """Validate that current hazen version satisfies the constraint.

        Raises:
            RuntimeError: If version constraint is not satisfied or invalid.

        """
        try:
            specifier = SpecifierSet(constraint_str)
        except InvalidSpecifier as e:
            msg = (
                f"Invalid hazen_version_constraint '{constraint_str}': {e}. "
                "Use standard semver specifiers like"
                " '>=2.0.0', '~=2.0.0', '==2.0.*' etc."
            )
            raise ValueError(msg) from e

        current_version = packaging_version.parse(__version__)

        if current_version not in specifier:
            msg = (
                "Version constraint mismatch:"
                f" Config requires hazen {constraint_str},"
                f" but you are running {__version__}."
                " Please install a compatible version:\n"
                f"pip install 'hazen{constraint_str}'"
                f" or\nuv tool install 'hazen{constraint_str}'"
            )
            raise RuntimeError(msg)

    @staticmethod
    def _check_and_set_default_report(
        levels: list[str],
        *,
        report: bool,
    ) -> bool:
        if "all" in levels or "intermediate" in levels:
            logger.info(
                "levels = %s requires report=True as a default"
                " so this has been set.",
                levels,
            )
            return True
        return report

    @classmethod
    def from_config(
        cls,
        config_path: str | Path,
        *,
        dry_run: bool = False,
    ) -> BatchConfig:
        """Read a configuration file into a BatchConfig object."""
        config_path = Path(config_path)
        with config_path.open("r") as f:
            data = yaml.safe_load(f)

        if not isinstance(data, dict):
            msg = f"{config_path} is not a properly formatted yaml file"
            raise TypeError(msg)

        constraint_str = data.get("hazen_version_constraint")
        if constraint_str:
            cls._validate_hazen_version(constraint_str)

        schema_version = data.get("version", "1.0")
        data = cls._validate_schema_version(schema_version, data)

        # Resolve paths relative to config file location
        config_dir = config_path.parent

        def resolve_path(path: str | Path) -> Path:
            p = Path(path)
            if p.is_absolute():
                return p
            # Always resolve relative paths with respect to the config file directory
            return (config_dir / p).absolute()

        # Parse jobs
        jobs = []
        for job_data in data.get("jobs", []):
            # Resolve folder paths
            folders = [resolve_path(f) for f in job_data.get("folders", [])]

            # Validate task name against registry
            task_name = job_data["task"]

            jobs.append(
                JobTaskConfig(
                    task=task_name,
                    folders=folders,
                    overrides=job_data.get("overrides", {}),
                ),
            )

        # Handle optional paths
        def resolve_path_as_str(path: Path | None) -> str | None:
            return resolve_path(path).as_posix() if path is not None else None

        report_docx = resolve_path_as_str(data.get("report_docx"))
        report_template = resolve_path_as_str(data.get("report_template"))
        output = resolve_path_as_str(data.get("output"))

        logger.info(
            "Batch config job created from: %s (hash: %s)",
            config_path,
            hashlib.sha256(str(data).encode("utf-8")).hexdigest(),
        )

        levels = data.get("levels", ["final"])
        if isinstance(levels, str):
            levels = [levels]

        defaults = data.get("defaults", {})
        defaults["report"] = cls._check_and_set_default_report(
            levels,
            report=defaults.get("report", False),
        )
        return cls(
            version=data.get("version", "1.0"),
            hazen_version_constraint=data.get("hazen_version_constraint"),
            description=data.get("description", ""),
            output=output,
            jobs=jobs,
            levels=levels,
            report_docx=report_docx,
            report_template=report_template,
            defaults=defaults,
            _file=config_path,
            _dry_run=dry_run,
        )


def _execute_task(
    task: str,
    files: list[str],
    kwargs: dict[str, Any],
) -> Result:
    """Encapsulate the work for a single task."""
    report = kwargs.pop("report", False)
    report_dir = kwargs.pop("report_dir", None)
    task = init_task(
        task,
        files,
        report=report,
        report_dir=report_dir,
        **kwargs,
    )
    return task.run()


##############
# Registries #
##############

PROTOCOL_REGISTRY: dict[str, type[Protocol]] = {
    "acr_all": ACRLargePhantomProtocol,
}

# Note that if changing the TASK_REGISTRY keys you should
# update Protocol to match up with the task registry.
TASK_REGISTRY = {
    # MagNET #
    "snr": TaskMetadata(
        module_name="snr",
        class_name="SNR",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "ghosting": TaskMetadata(
        module_name="ghosting",
        class_name="Ghosting",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "uniformity": TaskMetadata(
        module_name="uniformity",
        class_name="Uniformity",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "spatial_resolution": TaskMetadata(
        module_name="spatial_resolution",
        class_name="SpatialResolution",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "slice_width": TaskMetadata(
        module_name="slice_width",
        class_name="SliceWidth",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "slice_position": TaskMetadata(
        module_name="slice_position",
        class_name="SlicePosition",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    "snr_map": TaskMetadata(
        module_name="snr_map",
        class_name="SNRMap",
        single_image=True,
        phantom=PhantomType.MAGNET,
    ),
    # ACR #
    "acr_geometric_accuracy": TaskMetadata(
        module_name="acr_geometric_accuracy",
        class_name="ACRGeometricAccuracy",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_ghosting": TaskMetadata(
        module_name="acr_ghosting",
        class_name="ACRGhosting",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_low_contrast_object_detectability": TaskMetadata(
        module_name="acr_low_contrast_object_detectability",
        class_name="ACRLowContrastObjectDetectability",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_object_detectability": TaskMetadata(
        module_name="acr_object_detectability",
        class_name="ACRObjectDetectability",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_slice_position": TaskMetadata(
        module_name="acr_slice_position",
        class_name="ACRSlicePosition",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_slice_thickness": TaskMetadata(
        module_name="acr_slice_thickness",
        class_name="ACRSliceThickness",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_snr": TaskMetadata(
        module_name="acr_snr",
        class_name="ACRSNR",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_spatial_resolution": TaskMetadata(
        module_name="acr_spatial_resolution",
        class_name="ACRSpatialResolution",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_sagittal_geometric_accuracy": TaskMetadata(
        module_name="acr_sagittal_geometric_accuracy",
        class_name="ACRSagittalGeometricAccuracy",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    "acr_uniformity": TaskMetadata(
        module_name="acr_uniformity",
        class_name="ACRUniformity",
        single_image=False,
        phantom=PhantomType.ACR,
    ),
    # Caliber
    "relaxometry": TaskMetadata(
        module_name="relaxometry",
        class_name="Relaxometry",
        single_image=False,
        phantom=PhantomType.CALIBER,
    ),
}



import unittest
import pathlib
import numpy as np

from hazenlib.utils import get_dicom_files
from hazenlib.tasks.hazen_spatial_resolution import ACRSpatialResolution
from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestACRSpatialResolutionSiemens(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensMTF")
    centre = (128, 124)
    rotation_angle = 9
    y_ramp_pos = 118
    width = 13
    edge_type = "vertical", "downward"
    edge_loc = [5, 7]
    slope = -0.165
    MTF50 = (1.16, 1.35)

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)

        self.acr_spatial_resolution_task = ACRSpatialResolution(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

        self.dcm = self.acr_spatial_resolution_task.ACR_obj.slice_stack[0]
        self.crop_image = self.acr_spatial_resolution_task.crop_image(
            self.dcm.pixel_array, self.centre[0], self.y_ramp_pos, self.width
        )
        self.data = self.dcm.pixel_array
        self.res = self.dcm.PixelSpacing

    def test_find_y_ramp(self):
        y_ramp_pos = self.acr_spatial_resolution_task.y_position_for_ramp(
            self.data, self.centre
        )
        assert y_ramp_pos == self.y_ramp_pos

    def test_get_edge_type(self):
        edge_type = self.acr_spatial_resolution_task.get_edge_type(
            self.crop_image
        )
        assert edge_type == self.edge_type

    def test_get_edge_loc(self):
        assert (
            self.acr_spatial_resolution_task.edge_location_for_plot(
                self.crop_image, self.edge_type[0] == self.edge_loc
            )
        ).all()

    def test_retrieve_slope(self):
        slope = self.acr_spatial_resolution_task.fit_normcdf_surface(
            self.crop_image, self.edge_type[0], self.edge_type[1]
        )[0]
        assert np.round(slope, 3) == self.slope

    def test_get_MTF50(self):
        mtf50 = self.acr_spatial_resolution_task.get_mtf50(self.dcm)
        rounded_mtf50 = (np.round(mtf50[0], 2), np.round(mtf50[1], 2))

        print("\ntest_get_MTF50.py::TestGetMTF50::test_get_MTF50")
        print("new_release_value:", rounded_mtf50)
        print("fixed_value:", self.MTF50)

        assert rounded_mtf50 == self.MTF50


class TestACRSpatialResolutionGE(TestACRSpatialResolutionSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    centre = (254, 255)
    rotation_angle = 0
    y_ramp_pos = 244
    width = 26
    edge_type = "vertical", "upward"
    edge_loc = [5, 7]
    slope = 0.037
    MTF50 = (0.72, 0.71)



"""Module to test the ACR sagittal localiser."""

# ruff: noqa:PT009

import logging
import pathlib
import unittest

import numpy as np
from hazenlib.tasks.acr_sagittal_geometric_accuracy import (
    ACRSagittalGeometricAccuracy,
)
from hazenlib.utils import get_dicom_files

from tests import TEST_DATA_DIR, TEST_REPORT_DIR

logger = logging.getLogger(__name__)


class TestACRSagittalGeometricAccuracySiemens(unittest.TestCase):
    """Base test class for the Siemens Sagittal Localizer Geometric Accuracy."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFitLocalizer")
    L1 = 149.41

    def setUp(self) -> None:
        """Set up the test case."""
        input_files = get_dicom_files(self.ACR_DATA)

        self.acr_geometric_accuracy_task = ACRSagittalGeometricAccuracy(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

        self.dcm_1 = self.acr_geometric_accuracy_task.ACR_obj.slice_stack[0]
        self.slice1_val = np.round(
            self.acr_geometric_accuracy_task.get_geometric_accuracy(
                self.dcm_1,
            ),
            2,
        )

    def test_geometric_accuracy_slice_1(self) -> None:
        """Test the geometric accuracy of slice 1."""
        logger.info(
            "new_release: %f\nfixed value: %f",
            self.slice1_val,
            self.L1,
        )
        self.assertAlmostEqual(self.slice1_val, self.L1, 0)


class TestACRSagittalGeometricAccuracyPhilipsAchieva(
    TestACRSagittalGeometricAccuracySiemens,
):
    """Base test class for the Philips Sagittal Localizer Geometric Accuracy."""

    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchievaLocalizer")
    L1 = 146.48


class TestACRSagittalGeometricAccuracyGESigna(
    TestACRSagittalGeometricAccuracySiemens,
):
    """Base test class for the GE Sagittal Localizer Geometric Accuracy."""

    ACR_DATA = (
        pathlib.Path(TEST_DATA_DIR)
        / "acr"
        / "GE_Signa_1.5T_Sagittal_Localizer"
    )
    L1 = 150.0



# Python imports
import unittest
import pathlib

# Module imports
import numpy as np

# Local imports
from hazenlib.utils import (
    get_dicom_files,
    create_circular_mask,
    create_circular_roi_at,
)
from hazenlib.tasks.acr_uniformity import ACRUniformity
from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestACRUniformitySiemens(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    piu = 67.49

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)

        self.acr_uniformity_task = ACRUniformity(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

    def test_roi_values_within_roi(self) -> None:
        """Test the ROI values actually lie within the ROI."""
        img = self.acr_uniformity_task.ACR_obj.slice_stack[6].pixel_array
        (
            (center_x, center_y),
            _,
        ) = self.acr_uniformity_task.ACR_obj.find_phantom_center(
            img,
            self.acr_uniformity_task.ACR_obj.dx,
            self.acr_uniformity_task.ACR_obj.dy,
        )

        large_roi = create_circular_roi_at(
            img,
            self.acr_uniformity_task.r_large,
            center_x,
            center_y,
        )

        large_roi_valid_space = create_circular_mask(
            img,
            self.acr_uniformity_task.r_large_filter,
            center_x,
            center_y,
        )

        (
            x_min,
            y_min,
            _,
            x_max,
            y_max,
            _,
        ) = self.acr_uniformity_task.get_mean_roi_values(
            large_roi,
            ~large_roi_valid_space,
        )

        for x, y in ([x_min, y_min], [x_max, y_max]):
            dist_from_center = (
                np.sqrt((x - center_x) ** 2 + (y - center_y) ** 2)
                + self.acr_uniformity_task.r_small
            )
            self.assertLessEqual(
                dist_from_center,
                self.acr_uniformity_task.r_large,
            )

    def test_uniformity(self):
        results = self.acr_uniformity_task.get_integral_uniformity(
            self.acr_uniformity_task.ACR_obj.slice_stack[6]
        )
        rounded_results = round(results, 2)

        print("\ntest_uniformity.py::TestUniformity::test_uniformity")
        print("new_release_values:", rounded_results)
        print("fixed_values:", self.piu)

        assert rounded_results == self.piu


class TestACRUniformityGESignaArtistT1(TestACRUniformitySiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR) / "acr" / "GE_Signa_Artist_1.5T_T1"
    piu = 92.03


class TestACRUniformitySiemensSolaFit(TestACRUniformitySiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "SiemensSolaFit")
    piu = 95.77


class TestACRUniformityPhilipsAchieva(TestACRUniformitySiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "PhilipsAchieva")
    piu = 80.66


class TestACRUniformityGE(TestACRUniformitySiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    piu = 83.78



"""Tests for output formatters."""

# ruff: noqa: D401 PT009 PT027

from __future__ import annotations

# Python imports
import io
import json
import tempfile
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

# Module imports
import numpy as np

# Local imports
from hazenlib.formatters import (
    _build_rows,
    _format_results,
    write_result,
)
from hazenlib.types import Measurement, Metadata, Result


class TestWriteResult(unittest.TestCase):
    """Unit tests for the write_result entry point."""

    def _create_result(self, visibility: str = "final") -> Result:
        """Helper to create a Result with measurements of specific visibility."""
        result = Result(task="SNR", desc="SNR test", files=["test.dcm"])
        result.metadata = Metadata(institution_name="Test Hospital")
        result.add_measurement(
            Measurement(
                name="SNR",
                value=42.5,
                type="measured",
                unit="dB",
                visibility=visibility,
            ),
        )
        return result

    def test_writes_json_to_stdout_by_default(self) -> None:
        """Verify JSON output is written to stdout when path is '-'."""
        result = self._create_result()

        with patch("sys.stdout", new=io.StringIO()) as mock_stdout:
            write_result(result, "json", path="-", level="all")
            output = mock_stdout.getvalue()

        # Verify structure
        data = json.loads(output)
        self.assertEqual(data["task"], "SNR")
        self.assertEqual(len(data["measurements"]), 1)
        self.assertEqual(data["measurements"][0]["value"], 42.5)

    def test_writes_csv_to_file_with_header(self) -> None:
        """Verify CSV is written to file with comma delimiter."""
        result = self._create_result()

        with tempfile.NamedTemporaryFile(
            mode="w",
            delete=False,
            suffix=".csv",
        ) as tmp:
            tmp_path = tmp.name

        try:
            write_result(result, "csv", path=tmp_path, level="all")

            with Path(tmp_path).open("r") as f:
                content = f.read()
                lines = content.strip().split("\n")

                # Header + 1 data row
                self.assertEqual(len(lines), 2)
                self.assertIn("name", lines[0])  # header
                self.assertIn("42.5", lines[1])  # data
                self.assertIn(",", content)  # comma delimited
        finally:
            Path(tmp_path).unlink(missing_ok=True)

    def test_level_filtering_excludes_intermediate(self) -> None:
        """Verify level='final' filters out intermediate measurements."""
        result = Result(task="SNR", desc="test", files=["1.dcm"])
        result.add_measurement(
            Measurement(
                name="SNR",
                subtype="final",
                value=1.0,
                visibility="final",
            ),
        )
        result.add_measurement(
            Measurement(
                name="SNR",
                subtype="intermediate",
                value=2.0,
                visibility="intermediate",
            ),
        )

        with patch("sys.stdout", new=io.StringIO()) as mock_stdout:
            write_result(result, "json", path="-", level="final")
            output = mock_stdout.getvalue()

        data = json.loads(output)
        self.assertEqual(len(data["measurements"]), 1)
        self.assertEqual(data["measurements"][0]["name"], "SNR")

    def test_invalid_level_defaults_to_all_with_warning(self) -> None:
        """Verify unknown level logs warning and defaults to 'all'."""
        result = self._create_result()

        with patch("hazenlib.formatters.logger") as mock_logger:
            with patch(
                "sys.stdout",
                new=io.StringIO(),
            ) as mock_stdout:  # noqa: F841
                write_result(result, "json", path="-", level="unknown_level")
                # Should not raise, should log warning

            mock_logger.warning.assert_called_once()
            self.assertIn(
                "Unknown measurement filter level",
                str(mock_logger.warning.call_args),
            )

    def test_invalid_format_raises_valueerror(self) -> None:
        """Verify ValueError is raised for unsupported format strings."""
        result = self._create_result()

        with self.assertRaises(ValueError) as context:
            write_result(result, "xml", path="-")  # type: ignore[arg-type]

        self.assertIn("Unrecognised format", str(context.exception))


class TestFormatResultsDispatcher(unittest.TestCase):
    """Tests for _format_results dispatching and formatting logic."""

    def test_json_format_calls_to_json(self) -> None:
        """Verify json format uses Result.to_json() method."""
        mock_result = Mock(spec=Result)
        mock_result.to_json.return_value = '{"task": "test"}'

        output = io.StringIO()
        _format_results(mock_result, "json", output)

        self.assertEqual(output.getvalue().strip(), '{"task": "test"}')
        mock_result.to_json.assert_called_once()

    def test_csv_uses_comma_delimiter(self) -> None:
        """Verify csv format uses comma as delimiter."""
        result = Result(task="Test", desc="test", files="file.dcm")
        result.add_measurement(
            Measurement(name="SNR", value=1.0, visibility="final"),
        )

        output = io.StringIO()
        _format_results(result, "csv", output, write_header=False)

        content = output.getvalue()
        lines = content.strip().split("\n")
        # Check first line contains commas, not tabs
        self.assertIn(",", lines[0])
        self.assertNotIn("\t", content)

    def test_tsv_uses_tab_delimiter(self) -> None:
        """Verify tsv format uses tab as delimiter."""
        result = Result(task="SNR", desc="test", files="file.dcm")
        result.add_measurement(
            Measurement(name="SNR", value=1.0, visibility="final"),
        )

        output = io.StringIO()
        _format_results(result, "tsv", output, write_header=False)

        content = output.getvalue()
        self.assertIn("\t", content)

    def test_csv_header_written_conditionally(self) -> None:
        """Verify write_header parameter controls header output."""
        result = Result(task="SNR", desc="test", files="file.dcm")
        result.add_measurement(
            Measurement(name="SNR", value=42.0, visibility="final"),
        )

        # With header (default)
        with_header = io.StringIO()
        _format_results(result, "csv", with_header, write_header=True)
        lines_with = with_header.getvalue().strip().split("\n")
        self.assertIn("name", lines_with[0])

        # Without header
        without_header = io.StringIO()
        _format_results(result, "csv", without_header, write_header=False)
        lines_without = without_header.getvalue().strip().split("\n")
        # Should still have content but no header row with field names
        self.assertEqual(len(lines_without) + 1, len(lines_with))


class TestBuildRows(unittest.TestCase):
    """Tests for _build_rows data transformation."""

    def test_flattens_nested_structure(self) -> None:
        """Verify nested Result/Metadata/Measurements are flattened into rows."""
        result = Result(
            task="ACRSNR",
            desc="SNR analysis",
            files=["1.dcm", "2.dcm"],
        )
        result.metadata = Metadata(
            institution_name="Test Hospital",
            manufacturer="Siemens",
        )
        result.add_measurement(
            Measurement(
                name="SNR",
                value=125.5,
                type="measured",
                unit="ratio",
                visibility="final",
            ),
        )

        rows = _build_rows(result)

        self.assertEqual(len(rows), 1)
        row = rows[0]
        # Measurement fields
        self.assertEqual(row["name"], "SNR")
        self.assertEqual(row["value"], 125.5)
        self.assertEqual(row["unit"], "ratio")
        # Result fields
        self.assertEqual(row["task"], "ACRSNR")
        # Metadata fields
        self.assertEqual(row["institution_name"], "Test Hospital")
        self.assertEqual(row["manufacturer"], "Siemens")

    def test_handles_multiple_measurements(self) -> None:
        """Verify multiple measurements create multiple rows."""
        result = Result(
            task="SNR",
            desc="Geometric accuracy",
            files=["1.dcm"],
        )
        result.add_measurement(
            Measurement(
                name="SNR",
                description="x_dim",
                value=100.0,
                visibility="final",
            ),
        )
        result.add_measurement(
            Measurement(
                name="SNR",
                description="y_dim",
                value=100.1,
                visibility="final",
            ),
        )

        rows = _build_rows(result)

        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0]["description"], "x_dim")
        self.assertEqual(rows[1]["description"], "y_dim")
        # Both rows should inherit same task/metadata
        self.assertEqual(rows[0]["task"], "SNR")
        self.assertEqual(rows[1]["task"], "SNR")

    def test_returns_empty_list_for_no_measurements(self) -> None:
        """Verify empty measurements returns empty list (no rows)."""
        result = Result(task="Empty", desc="No data", files=[])
        rows = _build_rows(result)

        self.assertEqual(rows, [])

    def test_converts_numpy_scalars(self) -> None:
        """Verify numpy float64 etc. are converted to Python scalars."""
        mock_value = np.float64(42.0)

        result = Result(task="Test", desc="test", files="f.dcm")
        result.add_measurement(
            Measurement(
                name="SNR",
                value=mock_value,  # type: ignore[arg-type]
                visibility="final",
            ),
        )

        rows = _build_rows(result)
        self.assertEqual(rows[0]["value"], 42.0)
        self.assertIsInstance(rows[0]["value"], float)


if __name__ == "__main__":
    unittest.main()



import pathlib
import unittest

from hazenlib.tasks.acr_ghosting import ACRGhosting
from hazenlib.utils import get_dicom_files

from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestACRGhostingSiemens(unittest.TestCase):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "Siemens")
    psg = 0.035

    def setUp(self):
        input_files = get_dicom_files(self.ACR_DATA)
        self.acr_ghosting_task = ACRGhosting(
            input_data=input_files,
            report_dir=pathlib.PurePath.joinpath(TEST_REPORT_DIR),
        )

    def test_ghosting(self):
        ghosting_val = round(
            self.acr_ghosting_task.get_signal_ghosting(
                self.acr_ghosting_task.ACR_obj.slice_stack[6]
            ),
            3,
        )

        print("\ntest_ghosting.py::TestGhosting::test_ghosting")
        print("new_release_value:", ghosting_val)
        print("fixed_value:", self.psg)

        assert ghosting_val == self.psg


class TestACRGhostingGE(TestACRGhostingSiemens):
    ACR_DATA = pathlib.Path(TEST_DATA_DIR / "acr" / "GE")
    psg = 0.485



"""Welcome to the hazen Command Line Interface.

The following Tasks are available:
- ACR phantom:
acr_all | acr_snr | acr_slice_position | acr_slice_thickness |
acr_spatial_resolution | acr_uniformity | acr_ghosting | acr_geometric_accuracy |
acr_low_contrast_object_detectability
- MagNET Test Objects:
snr | snr_map | slice_position | slice_width | spatial_resolution | uniformity | ghosting
- Caliber phantom:
relaxometry

Note that the acr_all task requires 3 directories as arguments
(T1, T2 and Sagittal Localiser) whilst all other commands require
a single positional directory argument. That is:

hazen acr_all /path/to/T1 /path/to/T2 /path/to/SagittalLocaliser
"""

import argparse
import logging
import os
import shutil
from pathlib import Path
from typing import get_args

from hazenlib._version import __version__
from hazenlib.constants import MEASUREMENT_VISIBILITY
from hazenlib.discovery import generate_batch_config
from hazenlib.execution import timed_execution
from hazenlib.formatters import write_result
from hazenlib.logger import logger
from hazenlib.orchestration import (
    TASK_REGISTRY,
    ACRLargePhantomProtocol,
    BatchConfig,
    init_task,
)
from hazenlib.utils import get_dicom_files


def get_parser() -> argparse.ArgumentParser:
    """Return the argument parser."""
    ####################
    # Common arguments #
    ####################

    common_parser = argparse.ArgumentParser(add_help=False)

    common_parser.add_argument(
        "--verbose",
        action="store_true",
        help=(
            "Whether to provide additional metadata about the calculation "
            "in the result (slice position and relaxometry tasks)"
        ),
    )
    common_parser.add_argument(
        "--log",
        type=str,
        default="info",
        choices=["debug", "info", "warning", "error", "critical"],
        help=(
            "Set the level of logging based on severity. "
            'Available levels are "debug", "warning", "error", "critical", '
            'with "info" as default'
        ),
    )

    common_parser.add_argument(
        "--version",
        action="version",
        version=__version__,
    )

    ###############
    # Main parser #
    ###############

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[common_parser],
    )

    ###########################################
    # General options available for all tasks #
    ###########################################

    task_options_parser = argparse.ArgumentParser(
        add_help=False,
        parents=[common_parser],
    )

    task_options_parser.add_argument(
        "--profile",
        action="store_true",
        help="Include execution time metadata in results",
    )

    task_options_parser.add_argument(
        "--report",
        action="store_true",
        help="Whether to generate visualisation of the measurement steps",
    )
    task_options_parser.add_argument(
        "--report-docx",
        type=str,
        default=None,
        help="Path to save Word report (requires --report for images)",
    )
    task_options_parser.add_argument(
        "--report-template",
        type=str,
        default=None,
        help="Path to template to be used for --report-docx report.",
    )
    task_options_parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Provide a folder where report images are to be saved",
    )

    task_options_parser.add_argument(
        "--format",
        type=str,
        default="json",
        choices=["json", "csv", "tsv"],
        help="Output format for test results. Choices: json (default), csv or tsv",
    )
    task_options_parser.add_argument(
        "--result",
        type=str,
        default="-",
        help='Path to the results path. If "-", default, will write to stdout',
    )
    task_options_parser.add_argument(
        "--level",
        type=str,
        default="all",
        choices=[*get_args(MEASUREMENT_VISIBILITY), "all"],
        help=(
            "Filter results by visibility:"
            " 'final' (report-ready metrics),"
            " 'intermediate' (scientist verification),"
            " or 'all'."
        ),
    )

    ######################
    # Set up sub parsers #
    ######################

    subparsers = parser.add_subparsers(
        dest="command",
        required=True,
        help="Available commands",
    )

    ####################
    # Batch sub parser #
    ####################

    batch_parser = subparsers.add_parser(
        "batch",
        help="Execute task from config files.",
        parents=[common_parser],
    )
    batch_parser.add_argument(
        "config",
        type=str,
        nargs="?",
        default=None,
        help="Path to YAML configuration file (required unless --init is used)",
    )
    batch_parser.add_argument(
        "--init",
        type=str,
        default=None,
        metavar="DIR",
        help=(
            "Generate a batch config file"
            " from a directory of DICOM acquisitions"
        ),
    )
    batch_parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Output path for generated config file (used with --init)",
    )
    batch_parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Validate config and list jobs without executing",
    )

    ###################
    # Task sub parser #
    ###################
    for task_name in TASK_REGISTRY:
        task_parser = subparsers.add_parser(
            task_name,
            parents=[task_options_parser],
        )
        task_parser.add_argument(
            "folder",
            help="Path to folder containing DICOM files",
            nargs=1,
        )

        ########################
        # SNR Specific Options #
        ########################
        if task_name in ["snr", "acr_snr"]:
            task_parser.add_argument(
                "--measured_slice_width",
                type=float,
                default=None,
                help=(
                    "Provide a slice width to be used for SNR measurement, "
                    "by default it is parsed from the DICOM "
                    "(optional for acr_snr and snr)"
                ),
            )
            task_parser.add_argument(
                "--subtract",
                type=str,
                default=None,
                help=(
                    "Provide a second folder to calculate SNR by subtraction "
                    "for the ACR phantom (optional for acr_snr)"
                ),
            )
            task_parser.add_argument(
                "--coil",
                type=str,
                default=None,
                choices=["head", "body"],
                help="Coil type for SNR measurement (optional for snr)",
            )

        ################################
        # Relaxometry specific options #
        ################################

        if task_name == "relaxometry":
            task_parser.add_argument(
                "--calc",
                type=str,
                default=None,
                choices=["T1", "T2"],
                help=(
                    "Choose 'T1' or 'T2' for relaxometry measurement "
                    "(required for relaxometry)"
                ),
            )
            task_parser.add_argument(
                "--plate_number",
                type=int,
                default=None,
                choices=[4, 5],
                help="Which plate to use for measurement: 4 or 5",
            )

    # Replace this with a PROTOCOL_REGISTRY similar to task registry
    for protocol_name in ["acr_all"]:
        protocol_parser = subparsers.add_parser(
            protocol_name,
            parents=[task_options_parser],
        )
        protocol_parser.add_argument(
            "folder",
            help="Path to folder(s) containing DICOM files",
            nargs="+",
        )

    return parser


def main() -> None:
    """Primary entrypoint to hazen."""
    parser = get_parser()
    args = parser.parse_args()

    try:
        execution_wrapper = (
            timed_execution
            if args.profile
            else (lambda f, *a, **k: f(*a, **k))
        )
    # Batch commands always run with timed execution.
    except AttributeError:
        execution_wrapper = timed_execution

    single_image_tasks = [
        task for task in TASK_REGISTRY.values() if task.single_image
    ]

    # Set common options
    log_levels = {
        "critical": logging.CRITICAL,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }
    level = log_levels.get(args.log, logging.INFO)
    logging.getLogger().setLevel(level)

    logger.info(f"Hazen version: {__version__}")

    ###################################
    # Special Case for batch commands #
    ###################################
    if args.command == "batch":
        # Init (discovery) command
        if args.init:
            init_path = Path(args.init)
            if not init_path.is_dir():
                parser.error(f"Directory not found: {init_path}")

            batch_config = generate_batch_config(init_path)

            # Override output file if specified
            if args.output:
                batch_config._file = Path(args.output)  # noqa: SLF001

            # Write the config file
            output_path = Path(batch_config._file)  # noqa: SLF001
            output_path.parent.mkdir(parents=True, exist_ok=True)
            batch_config.to_yaml(output_path)
            print(f"Batch config written to: {output_path}")  # noqa: T201
            return

        if not args.config:
            parser.error("config is required when not using --init")

        if not Path(args.config).exists():
            parser.error(f"Config file not found: {args.config}")
        batch = BatchConfig.from_config(args.config, dry_run=args.dry_run)

        output = Path(batch.output)
        output.parent.mkdir(parents=True, exist_ok=True)
        results = execution_wrapper(batch.run)

        if args.dry_run:
            return

        if batch.report_docx:
            for level in batch.levels:
                doc = results.to_docx(
                    template_path=batch.report_template,
                    level=level,
                )
                out = Path(batch.report_docx)
                doc.save(out.with_stem(out.stem + f"_{level}"))

        for level in batch.levels:
            for result in results.results:
                write_result(
                    result,
                    fmt=output.suffix.split(".")[-1],
                    path=output.with_stem(
                        output.stem + f"_{level}",
                    ),
                    level=level,
                )

        conf_src = Path(args.config)
        conf_bak = conf_src.with_suffix(conf_src.suffix + ".bak")
        shutil.copy(conf_src, conf_bak)
        print(  # noqa: T201
            "Batch job successfully run!"
            f" Current batch file copied to {conf_bak} as a backup.",
        )
        return

    #############################
    # Gathers non-batch options #
    #############################

    report = args.report
    report_dir = args.output
    verbose = args.verbose
    fmt = args.format
    level = args.level
    result_file = args.result

    # Parse the task and optional arguments:
    selected_task = args.command.lower()

    #################################
    # Special Case the ACR ALL Task #
    #################################

    if selected_task == "acr_all":
        task = ACRLargePhantomProtocol(
            args.folder,
            report=report,
            report_dir=report_dir,
            verbose=verbose,
        )
        protocol = execution_wrapper(task.run)
        if args.report_docx or args.report_template:
            doc = protocol.to_docx(
                template_path=args.report_template,
                level=level,
            )
            doc.save(args.report_docx)

        for result in protocol.results:
            write_result(result, fmt=fmt, path=result_file, level=level)
        return
    if len(args.folder) != 1:
        parser.error(
            f"Task '{selected_task}' expects exactly one folder"
            f" as a positional argument, but {len(args.folder)}"
            " were provided",
        )

    #####################
    # Single task usage #
    #####################

    logger.debug("The following files were identified as valid DICOMs:")
    files = get_dicom_files(args.folder[0])
    logger.debug(
        "%s task will be set off on %s images",
        args.command,
        len(files),
    )

    if selected_task == "snr":
        task = init_task(
            selected_task,
            files,
            report,
            report_dir,
            measured_slice_width=args.measured_slice_width,
            coil=args.coil,
        )
        result = execution_wrapper(task.run)
    elif selected_task == "acr_snr":
        task = init_task(
            selected_task,
            files,
            report,
            report_dir,
            subtract=args.subtract,
            measured_slice_width=args.measured_slice_width,
        )
        result = execution_wrapper(task.run)
    elif selected_task == "relaxometry":
        missing_args = []
        if args.calc is None:
            missing_args.append("--calc")
        if args.plate_number is None:
            missing_args.append("--plate_number")
        if missing_args:
            parser.error(
                f"relaxometry task requires the following arguments: "
                f"{', '.join(missing_args)}",
            )
        task = init_task(selected_task, files, report, report_dir)
        result = execution_wrapper(
            task.run,
            calc=args.calc,
            plate_number=args.plate_number,
            verbose=verbose,
        )
    else:
        if selected_task in single_image_tasks:
            # Ghosting, Uniformity, Spatial resolution, SNR map, Slice width
            # for now these are most likely not enhanced, single-frame
            for f in files:
                task = init_task(selected_task, [f], report, report_dir)
                result = execution_wrapper(task.run)
                write_result(result, fmt=fmt, path=result_file, level=level)
            return
        # Slice Position task, all ACR tasks except SNR
        # may be enhanced, may be multi-frame
        fns = [os.path.basename(fn) for fn in files]
        logger.info("Processing: %s", fns)
        task = init_task(
            selected_task,
            files,
            report,
            report_dir,
            verbose=verbose,
        )
        result = execution_wrapper(task.run)

        write_result(result, fmt=fmt, path=result_file, level=level)
        return

    write_result(result, fmt=fmt, path=result_file, level=level)


if __name__ == "__main__":
    main()



"""Application-specific errors"""

# Python imports
from collections.abc import Sequence

# Local imports
from hazenlib.constants import (
    MEASUREMENT_NAMES,
    MEASUREMENT_TYPES,
    MEASUREMENT_VISIBILITY,
)


class ShapeError(Exception):
    """Base exception for shapes."""

    def __init__(self, shape, msg=None):
        if msg is None:
            # Default message
            msg = f"An error occured with {shape}"
        self.msg = msg
        self.shape = shape


class ShapeDetectionError(ShapeError):
    """Shape not found"""

    def __init__(self, shape, msg=None):
        if msg is None:
            # Default message
            msg = f"Could not find shape: {shape}"
        super(ShapeError, self).__init__(msg)
        self.shape = shape


class MultipleShapesError(ShapeDetectionError):
    """Shape not found"""

    def __init__(self, shape, msg=None):
        if msg is None:
            # Default message
            msg = f"Multiple {shape}s found. Multiple shape detection is currently unsupported."

        super(ShapeDetectionError, self).__init__(msg)
        self.shape = shape


class ArgumentCombinationError(Exception):
    """Argument combination not valid."""

    def __init__(self, msg="Invalid combination of arguments."):
        super().__init__(msg)


class InvalidMeasurementNameError(ValueError):
    """Invalid Measurement Name Error."""

    def __init__(self, name: str) -> None:
        """Initialise the error."""
        msg = (
            f"Invalid measurement name: {name}."
            f" Must be one of {MEASUREMENT_NAMES}"
        )
        super().__init__(msg)


class InvalidMeasurementTypeError(ValueError):
    """Invalid Measurement Type Error."""

    def __init__(self, measurement_type: str) -> None:
        """Initialise the error."""
        msg = (
            f"Invalid measurement type: {measurement_type}."
            f" Must be one of {MEASUREMENT_TYPES}"
        )
        super().__init__(msg)


class InvalidMeasurementVisibilityError(ValueError):
    """Invalid Measurement Visibility Error."""

    def __init__(self, measurement_visibility: str) -> None:
        """Initialise the error."""
        msg = (
            f"Invalid measurement visibility: {measurement_visibility}."
            f" Must be one of {MEASUREMENT_VISIBILITY}"
        )
        super().__init__(msg)


class UnknownAcquisitionTypeError(ValueError):
    """Unknown Acquisition Type Error."""

    def __init__(self, acquisition_type: str) -> None:
        """Initialise the error."""
        msg = f"Unknown acquisition type: {acquisition_type}"
        super().__init__(msg)


class UnknownTaskNameError(ValueError):
    """Unknown Task Name Error."""

    def __init__(
        self,
        task_name: str,
        available_tasks: str | Sequence[str] | None,
    ) -> None:
        """Initialise the error."""
        msg = f"Unknown task '{task_name}'"
        if available_tasks is not None:
            msg += f". Available tasks: {available_tasks}"

        super().__init__(msg)



"""Tests for the orchestration module."""

# ruff: noqa: PT009 PT027 SLF001 S108

# Python imports
import shutil
import tempfile
import unittest
from collections.abc import Callable
from pathlib import Path
from unittest.mock import Mock, call, mock_open, patch

# Module imports
import yaml

# Local imports
from hazenlib.exceptions import (
    UnknownAcquisitionTypeError,
    UnknownTaskNameError,
)
from hazenlib.HazenTask import HazenTask
from hazenlib.orchestration import (
    PROTOCOL_REGISTRY,
    TASK_REGISTRY,
    AcquisitionType,
    ACRLargePhantomProtocol,
    BatchConfig,
    JobTaskConfig,
    Protocol,
    ProtocolResult,
    ProtocolStep,
    init_task,
)
from hazenlib.types import Result
from tests import TEST_DATA_DIR, TEST_REPORT_DIR


class TestInitTask(unittest.TestCase):
    """Unit tests for task initialization."""

    def test_successful_initialization(self) -> None:
        """Verify successful task creation from registry."""
        # Arrange
        mock_task_class = Mock(spec=HazenTask)
        mock_module = Mock()
        mock_module.SNR = mock_task_class

        with patch(
            "hazenlib.orchestration.importlib.import_module",
            return_value=mock_module,
        ):
            # Act
            args = (
                "snr",
                ["file1.dcm", "file2.dcm"],
            )
            kwargs = {
                "report": True,
                "report_dir": TEST_REPORT_DIR,
            }
            result = init_task(
                *args,
                **kwargs,
            )

            # Assert
            self.assertEqual(result, mock_task_class.return_value)
            mock_task_class.assert_called_once_with(
                input_data=args[1],
                **kwargs,
            )

    def test_unknown_task_raises_value_error(self) -> None:
        """Verify ValueError raised for unknown task names."""
        with self.assertRaises(ValueError) as context:
            init_task(
                "unknown_task",
                [],
                report=False,
                report_dir=TEST_REPORT_DIR,
            )

        self.assertIn("Unknown task", str(context.exception))

    def test_missing_class_raises_import_error(self) -> None:
        """Verify ImportError when module lacks expected class."""
        # Mock module without the expected class attribute
        mock_module = Mock(spec=["__name__", "NotSNR"])

        with (
            patch(
                "hazenlib.orchestration.importlib.import_module",
                return_value=mock_module,
            ),
            self.assertRaises(ImportError) as context,
        ):
            init_task("snr", [], report=False, report_dir=TEST_REPORT_DIR)

        self.assertIn("has no class", str(context.exception))


class TestAcquisitionType(unittest.TestCase):
    """Unit tests for AcquisitionType enum."""

    def test_from_string_t1(self) -> None:
        """Verify T1 string parsing."""
        result = AcquisitionType.from_string("t1")
        self.assertEqual(result, AcquisitionType.ACR_T1)
        # Case insensitive
        result = AcquisitionType.from_string("T1")
        self.assertEqual(result, AcquisitionType.ACR_T1)

    def test_from_string_t2(self) -> None:
        """Verify T2 string parsing."""
        result = AcquisitionType.from_string("t2")
        self.assertEqual(result, AcquisitionType.ACR_T2)

    def test_from_string_sagittal_variants(self) -> None:
        """Verify sagittal localizer parsing with spelling variants."""
        # Full string
        result = AcquisitionType.from_string("sagittal localiser")
        self.assertEqual(result, AcquisitionType.ACR_SL)

        # American spelling
        result = AcquisitionType.from_string("sagittal localizer")
        self.assertEqual(result, AcquisitionType.ACR_SL)

        # Abbreviations
        result = AcquisitionType.from_string("sagittal")
        self.assertEqual(result, AcquisitionType.ACR_SL)

        result = AcquisitionType.from_string("localizer")
        self.assertEqual(result, AcquisitionType.ACR_SL)

    def test_from_string_invalid_raises_error(self) -> None:
        """Verify UnknownAcquisitionTypeError for invalid strings."""
        with self.assertRaises(UnknownAcquisitionTypeError):
            AcquisitionType.from_string("unknown_type")

        with self.assertRaises(UnknownAcquisitionTypeError):
            AcquisitionType.from_string("t3")


class TestProtocolStep(unittest.TestCase):
    """Unit tests for ProtocolStep dataclass."""

    def test_valid_initialization(self) -> None:
        """Verify ProtocolStep can be created with valid task name."""
        step = ProtocolStep("snr", AcquisitionType.ACR_T1)
        self.assertEqual(step.task_name, "snr")
        self.assertEqual(step.acquisition_type, AcquisitionType.ACR_T1)

    def test_invalid_task_name_raises_error(self) -> None:
        """Verify UnknownTaskNameError raised for unregistered tasks."""
        with self.assertRaises(UnknownTaskNameError):
            ProtocolStep("nonexistent_task", AcquisitionType.ACR_T1)


class TestProtocol(unittest.TestCase):
    """Unit tests for Protocol dataclass."""

    def test_basic_initialization(self) -> None:
        """Verify Protocol can be initialized with name and steps."""
        steps = (
            ProtocolStep("snr", AcquisitionType.ACR_T1),
            ProtocolStep("ghosting", AcquisitionType.ACR_T1),
        )
        protocol = Protocol(name="Test Protocol", steps=steps)
        self.assertEqual(protocol.name, "Test Protocol")
        self.assertEqual(len(protocol.steps), 2)

    def test_empty_steps_default(self) -> None:
        """Verify Protocol defaults to empty steps tuple."""
        protocol = Protocol(name="Empty Protocol")
        self.assertEqual(protocol.steps, ())

    def test_from_config_not_implemented(self) -> None:
        """Verify from_config raises NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            Protocol.from_config(TEST_DATA_DIR / "config.json")


class TestProtocolResult(unittest.TestCase):
    """Unit tests for ProtocolResult class."""

    def test_initialization(self) -> None:
        """Verify ProtocolResult initialization."""
        result = ProtocolResult(task="TestProtocol", desc="test description")
        self.assertEqual(result.task, "TestProtocol")
        self.assertEqual(result.desc, "test description")
        self.assertEqual(len(result.results), 1)  # Initial result

    def test_add_result(self) -> None:
        """Verify results can be added to collection."""
        protocol_result = ProtocolResult(task="Protocol", desc="test")
        mock_result = Result(task="SubTask", desc="subtask result")

        protocol_result.add_result(mock_result)

        self.assertEqual(len(protocol_result.results), 2)
        self.assertEqual(protocol_result.results[-1], mock_result)

    def test_results_immutable(self) -> None:
        """Verify results property returns immutable tuple."""
        protocol_result = ProtocolResult(task="Protocol", desc="test")
        protocol_result.add_result(Result(task="Task1", desc="desc1"))

        results = protocol_result.results
        self.assertIsInstance(results, tuple)

        # Verify immutability
        with self.assertRaises(TypeError):
            results[0] = Result(task="Task2", desc="desc2")

    def test_add_multiple_results(self) -> None:
        """Verify multiple results can be added and retrieved."""
        protocol_result = ProtocolResult(task="Protocol", desc="test")
        num_results = 3
        results = [
            Result(task=f"Task{i}", desc=f"result{i}")
            for i in range(num_results)
        ]

        for r in results:
            protocol_result.add_result(r)

        # Protocol results always start off with an initial result
        # for the result of the protocol.
        self.assertEqual(len(protocol_result.results), num_results + 1)
        for i, r in enumerate(protocol_result.results[1:]):
            self.assertEqual(r.task, f"Task{i}")


class TestProtocolResultToDocx(unittest.TestCase):
    """Unit tests for ProtocolResult.to_docx method."""

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    def test_empty_results_raises_error(
        self,
        mock_inches: Mock,  # noqa: ARG002
        mock_document: Mock,  # noqa: ARG002
    ) -> None:
        """Verify ValueError raised when results list is empty."""
        protocol_result = ProtocolResult(task="Protocol", desc="test")
        # Access private attribute to simulate uninitialized state
        protocol_result._results = []  # type: ignore[misc]

        with patch.object(ProtocolResult, "results", None):
            with self.assertRaises(ValueError) as context:
                protocol_result.to_docx()
            self.assertIn("Results cannot be empty", str(context.exception))

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    def test_creates_document_without_template(
        self,
        mock_inches: Mock,  # noqa: ARG002
        mock_document_class: Mock,
    ) -> None:
        """Verify Document created and protocol header added when no template."""
        mock_doc = Mock()
        mock_document_class.return_value = mock_doc
        protocol_result = ProtocolResult(
            task="MyProtocol",
            desc="test description",
        )

        result = protocol_result.to_docx()

        self.assertEqual(result, mock_doc)
        mock_document_class.assert_called_once_with()
        mock_doc.add_heading.assert_called_once_with(
            "QA Report: MyProtocol",
            0,
        )

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    @patch("pathlib.Path.exists")
    def test_uses_template_when_provided(
        self,
        mock_exists: Mock,
        mock_inches: Mock,  # noqa: ARG002
        mock_document_class: Mock,
    ) -> None:
        """Verify template document is loaded when path provided."""
        mock_exists.return_value = True

        mock_doc = Mock()
        mock_document_class.return_value = mock_doc
        protocol_result = ProtocolResult(task="Protocol", desc="test")
        template_path = Path("/path/to/template.docx")

        result = protocol_result.to_docx(template_path=template_path)

        self.assertEqual(result, mock_doc)
        mock_document_class.assert_called_once_with(template_path)
        # Should not add protocol header when using template
        protocol_header_calls = [
            c
            for c in mock_doc.add_heading.call_args_list
            if c == call("QA Report: Protocol", 0)
        ]
        self.assertEqual(len(protocol_header_calls), 0)

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    def test_skips_protocol_metadata_result(
        self,
        mock_inches: Mock,  # noqa: ARG002
        mock_document_class: Mock,
    ) -> None:
        """Verify result matching protocol task name is skipped in output."""
        mock_doc = Mock()
        mock_document_class.return_value = mock_doc
        protocol_result = ProtocolResult(task="Protocol", desc="test")

        # Result with same task name as protocol (should be skipped)
        same_name_result = Mock(spec=Result)
        same_name_result.task = "Protocol"

        # Result with different task name (should be included)
        subtask_result = Mock(spec=Result)
        subtask_result.task = "SubTask"
        subtask_result.filtered.return_value = subtask_result
        subtask_result.measurements = []
        subtask_result.report_images = []

        protocol_result.add_result(same_name_result)
        protocol_result.add_result(subtask_result)

        mock_table = Mock()
        mock_doc.add_table.return_value = mock_table
        mock_row_mock = Mock()
        mock_row_mock.cells = [Mock() for _ in range(6)]
        mock_table.rows = [mock_row_mock]
        mock_new_row = Mock()
        mock_new_row.cells = [Mock() for _ in range(6)]
        mock_table.add_row.return_value = mock_new_row

        protocol_result.to_docx()

        # Check headings: protocol header (0) + one task (1), but not metadata
        heading_calls = mock_doc.add_heading.call_args_list
        self.assertEqual(len(heading_calls), 2)
        self.assertEqual(heading_calls[0], call("QA Report: Protocol", 0))
        self.assertEqual(heading_calls[1], call("SubTask", level=1))

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    def test_applies_level_filter_to_results(
        self,
        mock_inches: Mock,  # noqa: ARG002
        mock_document_class: Mock,
    ) -> None:
        """Verify level parameter passed to result"s filtered method."""
        mock_doc = Mock()
        mock_document_class.return_value = mock_doc
        protocol_result = ProtocolResult(task="Protocol", desc="test")

        mock_result = Mock(spec=Result)
        mock_result.task = "Task"
        mock_result.filtered.return_value = mock_result
        mock_result.measurements = []
        mock_result.report_images = []

        protocol_result.add_result(mock_result)

        mock_table = Mock()
        mock_doc.add_table.return_value = mock_table
        mock_row_mock = Mock()
        mock_row_mock.cells = [Mock() for _ in range(6)]
        mock_table.rows = [mock_row_mock]

        mock_new_row = Mock()
        mock_new_row.cells = [Mock() for _ in range(6)]
        mock_table.add_row.return_value = mock_new_row

        protocol_result.to_docx(level="final")

        mock_result.filtered.assert_called_once_with("final")

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    def test_creates_table_with_correct_structure(
        self,
        mock_inches: Mock,  # noqa: ARG002
        mock_document_class: Mock,
    ) -> None:
        """Verify measurements table has 6 columns and correct headers."""
        mock_doc = Mock()
        mock_table = Mock()
        # Mock header cells (6 columns)
        mock_hdr_cells = [Mock() for _ in range(6)]
        mock_table.rows = [Mock(cells=mock_hdr_cells)]
        # Mock row cells for measurement data
        mock_row_cells = [Mock() for _ in range(6)]
        mock_table.add_row.return_value = Mock(cells=mock_row_cells)
        mock_doc.add_table.return_value = mock_table
        mock_document_class.return_value = mock_doc

        protocol_result = ProtocolResult(task="Protocol", desc="test")

        # Create mock measurement
        mock_measurement = Mock()
        mock_measurement.name = "snr"
        mock_measurement.type = "measured"
        mock_measurement.subtype = ""
        mock_measurement.description = "Signal to noise"
        mock_measurement.value = 42.5
        mock_measurement.unit = "ratio"

        mock_result = Mock(spec=Result)
        mock_result.task = "Task"
        mock_result.filtered.return_value = mock_result
        mock_result.measurements = [mock_measurement]
        mock_result.report_images = []

        protocol_result.add_result(mock_result)
        protocol_result.to_docx()

        # Verify table created with correct dimensions and style
        mock_doc.add_table.assert_called_once_with(rows=1, cols=6)
        self.assertEqual(mock_table.style, "Light Grid Accent 1")

        # Verify headers populated
        expected_headers = [
            "Name",
            "Type",
            "Subtype",
            "Description",
            "Value",
            "Unit",
        ]
        for i, header in enumerate(expected_headers):
            self.assertEqual(mock_hdr_cells[i].text, header)

        # Verify measurement data converted to strings
        mock_table.add_row.assert_called_once()
        self.assertEqual(mock_row_cells[0].text, "snr")
        self.assertEqual(mock_row_cells[1].text, "measured")
        self.assertEqual(mock_row_cells[4].text, "42.5")

    @patch("hazenlib.orchestration.Document")
    @patch("hazenlib.orchestration.Inches")
    @patch("pathlib.Path.exists")
    def test_adds_report_images_with_five_inch_width(
        self,
        mock_exists: Mock,
        mock_inches: Mock,
        mock_document_class: Mock,
    ) -> None:
        """Verify report images embedded with 5.0 inch width."""
        mock_exists.return_value = True

        mock_doc = Mock()
        mock_document_class.return_value = mock_doc
        mock_inches.return_value = "5_inches_width"

        protocol_result = ProtocolResult(task="Protocol", desc="test")

        mock_result = Mock(spec=Result)
        mock_result.task = "Task"
        mock_result.filtered.return_value = mock_result
        mock_result.measurements = []
        mock_result.report_images = [
            "/path/to/image1.png",
            "/path/to/image2.png",
        ]

        protocol_result.add_result(mock_result)

        mock_table = Mock()
        mock_doc.add_table.return_value = mock_table
        mock_row_mock = Mock()
        mock_row_mock.cells = [Mock() for _ in range(6)]
        mock_table.rows = [mock_row_mock]
        mock_new_row = Mock()
        mock_new_row.cells = [Mock() for _ in range(6)]
        mock_table.add_row.return_value = mock_new_row

        protocol_result.to_docx(level="all")

        self.assertEqual(mock_doc.add_picture.call_count, 2)
        mock_doc.add_picture.assert_any_call(
            "/path/to/image1.png",
            width="5_inches_width",
        )
        mock_doc.add_picture.assert_any_call(
            "/path/to/image2.png",
            width="5_inches_width",
        )
        mock_inches.assert_called_with(5.0)


class TestACRLargePhantomProtocol(unittest.TestCase):
    """Integration tests for ACRLargePhantomProtocol.

    Attributes:
        PROTOCOL_STEPS: Number of steps for the protocol.
        MIN_DIRS : Minimum number of directories for the protocol.

    """

    PROTOCOL_STEPS: int = 15
    MIN_DIRS: int = 3

    def setUp(self) -> None:
        """Set up the test cases."""
        self.dirs = [
            TEST_DATA_DIR / "acr" / seq
            for seq in (
                "Siemens_Aera_1.5T_T1",
                "Siemens_Aera_1.5T_T2",
                "SiemensSolaFitLocalizer",
            )
        ]

    @patch("hazenlib.orchestration.ACRObject")
    def test_initialization_correct_dir_count(
        self,
        mock_acr_obj: Callable,
    ) -> None:
        """Verify protocol initializes with correct number of directories."""
        # Arrange
        mock_acr_instances = []
        for acq_type in ["T1", "T2", "sagittal localizer"]:
            mock_inst = Mock()
            mock_inst.acquisition_type.return_value = acq_type
            mock_acr_instances.append(mock_inst)

        mock_acr_obj.side_effect = mock_acr_instances

        # Act
        protocol = ACRLargePhantomProtocol(dirs=self.dirs)

        # Assert
        self.assertEqual(protocol.name, "ACR Large Phantom")
        self.assertEqual(len(protocol.steps), self.PROTOCOL_STEPS)
        self.assertEqual(len(protocol.file_groups), self.MIN_DIRS)

    def test_initialization_wrong_dir_count_raises_error(self) -> None:
        """Verify ValueError raised when directory count mismatch."""
        # Arrange - only provide 2 dirs when 3 unique types required
        # that is, missing localizer.
        dirs = self.dirs[:-1]

        with self.assertRaises(ValueError) as context:
            ACRLargePhantomProtocol(dirs=dirs)

        self.assertIn(
            "Incorrect number of directories",
            str(context.exception),
        )

    def test_nonunique_sequences_raise_error(self) -> None:
        """Verify ValueError raised when duplicated sequences are passed."""
        dirs = [*self.dirs[:2], self.dirs[0]]

        with self.assertRaises(ValueError) as context:
            ACRLargePhantomProtocol(dirs=dirs)

        self.assertIn(
            "Missing sequences",
            str(context.exception),
        )

    @patch("hazenlib.orchestration.ACRObject")
    @patch("hazenlib.orchestration.init_task")
    def test_run_executes_all_steps(
        self,
        mock_init_task: Callable,
        mock_acr_obj: Callable,
    ) -> None:
        """Verify run() executes all protocol steps and returns results."""
        # Arrange
        mock_acr_instances = []
        for acq_type in ["T1", "T2", "sagittal localizer"]:
            mock_inst = Mock()
            mock_inst.acquisition_type.return_value = acq_type
            mock_acr_instances.append(mock_inst)

        mock_acr_obj.side_effect = mock_acr_instances

        mock_task = Mock()
        mock_task.run.return_value = Result(task="MockTask", desc="done")
        mock_init_task.return_value = mock_task

        protocol = ACRLargePhantomProtocol(dirs=self.dirs)

        # Act
        # Use debug=True to disable using worker processes
        # which aren't registered in call_count
        result = protocol.run(debug=True)

        # Assert
        self.assertIsInstance(result, ProtocolResult)
        # Should initialize task for each step
        self.assertEqual(mock_init_task.call_count, self.PROTOCOL_STEPS)
        self.assertEqual(len(result.results), self.PROTOCOL_STEPS + 1)

    def test_steps_contain_expected_tasks(self) -> None:
        """Verify default steps contain expected task names."""
        with (
            patch(
                "hazenlib.orchestration.get_dicom_files",
                return_value=["mock_dicom.dcm"],
            ),
            patch(
                "hazenlib.orchestration.ACRObject",
            ) as mock_acr,
            patch(
                "hazenlib.orchestration.dcmread",
            ) as mock_dcmread,
        ):
            mock_inst = Mock()
            mock_inst.acquisition_type.side_effect = [
                "T1",
                "T2",
                "sagittal localizer",
            ]
            mock_acr.return_value = mock_inst

            mock_dcmread.return_value = Mock()

            protocol = ACRLargePhantomProtocol(dirs=self.dirs)

            task_names = [s.task_name for s in protocol.steps]
            expected_tasks = [
                "acr_geometric_accuracy",
                "acr_sagittal_geometric_accuracy",
                "acr_spatial_resolution",
                "acr_slice_thickness",
                "acr_slice_position",
                "acr_uniformity",
                "acr_ghosting",
                "acr_low_contrast_object_detectability",
                "acr_snr",
            ]

            for task in expected_tasks:
                self.assertIn(task, task_names)


class TestJobTaskConfig(unittest.TestCase):
    """Unit tests for JobTaskConfig dataclass."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.test_folder = Path("/tmp/test_folder")
        self.existing_folder = Path("/existing/folder")
        self.nonexistent_folder = Path("/nonexistent/folder")

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_valid_task_in_task_registry(self, mock_exists: Mock) -> None:
        """Verify task in TASK_REGISTRY sets is_protocol=False."""
        mock_exists.return_value = True

        config = JobTaskConfig(
            task="test_task",
            folders=[self.existing_folder],
        )

        self.assertEqual(config.task, "test_task")
        self.assertFalse(config.is_protocol)
        self.assertEqual(config.overrides, {})

    @patch.dict(PROTOCOL_REGISTRY, {"test_protocol": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_valid_task_in_protocol_registry(self, mock_exists: Mock) -> None:
        """Verify task in PROTOCOL_REGISTRY sets is_protocol=True."""
        mock_exists.return_value = True

        config = JobTaskConfig(
            task="test_protocol",
            folders=[
                self.existing_folder,
                self.existing_folder,
                self.existing_folder,
            ],
        )

        self.assertTrue(config.is_protocol)

    def test_unknown_task_raises_error(self) -> None:
        """Verify UnknownTaskNameError for unregistered tasks."""
        with self.assertRaises(UnknownTaskNameError) as context:
            JobTaskConfig(
                task="unknown_task",
                folders=[self.existing_folder],
            )

        self.assertIn("Unknown task", str(context.exception))
        self.assertIn("unknown_task", str(context.exception))

    @patch.dict(
        PROTOCOL_REGISTRY,
        {"acr_all": Mock()},
        clear=False,
    )
    @patch("pathlib.Path.exists")
    def test_acr_all_protocol_validates_three_folders(
        self,
        mock_exists: Mock,
    ) -> None:
        """Verify acr_all protocol accepts exactly 3 folders."""
        mock_exists.return_value = True

        config = JobTaskConfig(
            task="acr_all",
            folders=[
                self.existing_folder,
                self.existing_folder,
                self.existing_folder,
            ],
        )

        self.assertTrue(config.is_protocol)

    @patch.dict(
        PROTOCOL_REGISTRY,
        {"acr_all": Mock()},
        clear=False,
    )
    def test_acr_all_with_wrong_folder_count_raises_error(self) -> None:
        """Verify ValueError when acr_all doesn't have exactly 3 folders."""
        with self.assertRaises(ValueError) as context:
            JobTaskConfig(
                task="acr_all",
                folders=[self.existing_folder, self.existing_folder],
            )

        self.assertIn("3 directories", str(context.exception))
        self.assertIn("got 2", str(context.exception))

    def test_nonexistent_folder_raises_file_not_found(self) -> None:
        """Verify FileNotFoundError when folder doesn't exist."""
        with (
            patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False),
            self.assertRaises(FileNotFoundError) as context,
        ):
            JobTaskConfig(
                task="test_task",
                folders=[self.nonexistent_folder],
            )

        self.assertIn("Folder not found", str(context.exception))

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_overrides_stored(self, mock_exists: Mock) -> None:
        """Verify overrides dict is stored correctly."""
        mock_exists.return_value = True
        overrides = {"param1": "value1", "param2": 42}

        config = JobTaskConfig(
            task="test_task",
            folders=[self.existing_folder],
            overrides=overrides,
        )

        self.assertEqual(config.overrides, overrides)


class TestBatchConfig(unittest.TestCase):
    """Unit tests for BatchConfig dataclass."""

    def setUp(self) -> None:
        """Set up test configuration data."""
        self.config_data = {
            "version": "1.0",
            "description": "Test batch config",
            "output": "/tmp/output",
            "jobs": [
                {
                    "task": "snr",
                    "folders": ["/tmp/data"],
                    "overrides": {"param": "value"},
                },
            ],
        }
        self.config_path = Path("/tmp/test_config.yaml")

    @patch("pathlib.Path.exists")
    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_loads_yaml(
        self,
        mock_yaml: Mock,
        mock_exists: Mock,
    ) -> None:
        """Verify YAML file is loaded and parsed."""
        mock_exists.return_value = True
        mock_yaml.return_value = self.config_data

        with patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False):
            config = BatchConfig.from_config(self.config_path)

        self.assertEqual(config.version, "1.0")
        self.assertEqual(config.description, "Test batch config")
        self.assertEqual(len(config.jobs), 1)

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_resolves_relative_paths(
        self,
        mock_yaml: Mock,
    ) -> None:
        """Verify relative paths are resolved relative to config file."""
        mock_yaml.return_value = {
            **self.config_data,
            "output": "relative_output",
            "jobs": [{"task": "snr", "folders": ["relative_folder"]}],
        }

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
        ):
            mock_exists.return_value = True
            config = BatchConfig.from_config("/home/user/config.yaml")

        # Check that relative paths were resolved
        start_str = Path("/home/user").absolute().as_posix()
        self.assertTrue(str(config.output).startswith(start_str))
        job_folders = config.jobs[0].folders
        self.assertTrue(job_folders[0].as_posix().startswith(start_str))

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_preserves_absolute_paths(
        self,
        mock_yaml: Mock,
    ) -> None:
        """Verify absolute paths are not modified."""
        mock_yaml.return_value = {
            **self.config_data,
            "output": "/absolute/output/path",
        }

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
        ):
            mock_exists.return_value = True
            config = BatchConfig.from_config("config.yaml")

        self.assertEqual(
            str(config.output),
            Path("/absolute/output/path").absolute().as_posix(),
        )

    def test_validate_hazen_version_valid_constraint(self) -> None:
        """Verify no error when version constraint satisfied."""
        # Test with broad constraint that should always pass
        with patch("hazenlib.orchestration.__version__", "2.0.0"):
            # Should not raise
            BatchConfig._validate_hazen_version(">=1.0.0")

    def test_validate_hazen_version_mismatch_raises_runtime_error(
        self,
    ) -> None:
        """Verify RuntimeError when version constraint not satisfied."""
        with (
            patch("hazenlib.orchestration.__version__", "1.0.0"),
            self.assertRaises(RuntimeError) as context,
        ):
            BatchConfig._validate_hazen_version(">=2.0.0")

        self.assertIn("Version constraint mismatch", str(context.exception))
        self.assertIn("pip install", str(context.exception))

    def test_validate_hazen_version_invalid_specifier(self) -> None:
        """Verify ValueError for invalid version specifier."""
        with self.assertRaises(ValueError) as context:
            BatchConfig._validate_hazen_version("not_a_valid_specifier")

        self.assertIn(
            "Invalid hazen_version_constraint",
            str(context.exception),
        )

    def test_validate_schema_version_current(self) -> None:
        """Verify no migration needed for current schema."""
        data = {"key": "value"}
        result = BatchConfig._validate_schema_version("1.0", data)
        self.assertEqual(result, data)

    def test_validate_schema_version_newer_raises_error(self) -> None:
        """Verify ValueError when config schema is newer than supported."""
        with self.assertRaises(ValueError) as context:
            BatchConfig._validate_schema_version("99.0", {})

        self.assertIn("newer than supported", str(context.exception))

    def test_migrate_config_not_implemented(self) -> None:
        """Verify NotImplementedError for schema migration."""
        with self.assertRaises(NotImplementedError):
            BatchConfig._migrate_config({}, "0.9")

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_validates_hazen_version(
        self,
        mock_yaml: Mock,
    ) -> None:
        """Verify hazen version constraint is validated during load."""
        mock_yaml.return_value = {
            **self.config_data,
            "hazen_version_constraint": ">=2.0.0",
        }

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.object(
                BatchConfig,
                "_validate_hazen_version",
            ) as mock_validate,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
        ):
            mock_exists.return_value = True
            BatchConfig.from_config(self.config_path)

            mock_validate.assert_called_once_with(">=2.0.0")

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_optional_report_paths(self, mock_yaml: Mock) -> None:
        """Verify optional report_docx and report_template paths handled."""
        mock_yaml.return_value = {
            **self.config_data,
            "report_docx": "report.docx",
            "report_template": "template.docx",
        }

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
        ):
            mock_exists.return_value = True
            config = BatchConfig.from_config("/tmp/config.yaml")

        self.assertEqual(
            config.report_docx,
            Path("/tmp/report.docx").absolute().as_posix(),
        )
        self.assertEqual(
            config.report_template,
            Path("/tmp/template.docx").absolute().as_posix(),
        )

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_from_config_default_values(self, mock_yaml: Mock) -> None:
        """Verify default values applied when not in config."""
        minimal_data = {
            "output": "/tmp/output",
            "jobs": [],
        }
        mock_yaml.return_value = minimal_data

        with patch("pathlib.Path.exists", return_value=True):
            config = BatchConfig.from_config(self.config_path)

        self.assertEqual(config.version, "1.0")  # default
        self.assertEqual(config.description, "")  # default
        self.assertEqual(config.levels, ["final"])  # default
        self.assertIsNone(config.report_docx)  # default
        self.assertEqual(config.defaults, {"report": False})  # default

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_run_dry_run_mode(self, mock_yaml: Mock) -> None:
        """Verify dry run prints job info without executing."""
        mock_yaml.return_value = self.config_data

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
            patch("builtins.print") as mock_print,
            patch("hazenlib.orchestration.get_dicom_files") as mock_get_files,
        ):
            mock_exists.return_value = True
            mock_get_files.return_value = ["file1.dcm", "file2.dcm"]
            config = BatchConfig.from_config(self.config_path)
            result = config.run(dry_run=True)

        # Verify result returned without executing
        self.assertEqual(result.task, "Batch Configuration Job")

        # Verify dry run output printed
        print_calls = [str(call) for call in mock_print.call_args_list]
        self.assertTrue(
            any("Dry run complete" in c for c in print_calls),
        )

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_run_executes_protocols(self, mock_yaml: Mock) -> None:
        """Verify protocols are executed during run."""
        mock_yaml.return_value = {
            **self.config_data,
            "jobs": [{"task": "acr_all", "folders": ["/t1", "/t2", "/sl"]}],
        }

        mock_protocol_class = Mock()
        mock_protocol_instance = Mock()
        mock_protocol_instance.run.return_value = Mock(
            task="acr_all",
            results=[],
        )
        mock_protocol_class.return_value = mock_protocol_instance

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(
                PROTOCOL_REGISTRY,
                {"acr_all": mock_protocol_class},
                clear=False,
            ),
            patch("hazenlib.orchestration.get_dicom_files") as mock_get_files,
        ):
            mock_exists.return_value = True
            mock_get_files.return_value = ["file.dcm"]
            config = BatchConfig.from_config(self.config_path)
            config.run(dry_run=False)

        mock_protocol_class.assert_called_once()
        mock_protocol_instance.run.assert_called_once()

    @patch("pathlib.Path.open", mock_open(read_data="{}"))
    @patch("yaml.safe_load")
    def test_run_executes_tasks(self, mock_yaml: Mock) -> None:
        """Verify individual tasks are executed during run."""
        mock_yaml.return_value = self.config_data

        with (
            patch("pathlib.Path.exists") as mock_exists,
            patch.dict(TASK_REGISTRY, {"snr": Mock()}, clear=False),
            patch("hazenlib.orchestration.get_dicom_files") as mock_get_files,
            patch(
                "hazenlib.orchestration.wait_on_parallel_results",
            ) as mock_wait,
        ):
            mock_exists.return_value = True
            mock_get_files.return_value = ["file1.dcm", "file2.dcm"]
            mock_wait.return_value = []

            config = BatchConfig.from_config(self.config_path)
            config.run(dry_run=False)

            # Verify task jobs were prepared for parallel execution
            mock_wait.assert_called_once()
            call_args = mock_wait.call_args
            self.assertEqual(call_args[1]["debug"], False)


class TestBatchConfigToYaml(unittest.TestCase):
    """Unit tests for BatchConfig.to_yaml method."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.temp_dir = Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        """Clean up temp directory."""
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_writes_basic_config(self, mock_exists: Mock) -> None:
        """Verify basic YAML file is written correctly."""
        mock_exists.return_value = True
        output_path = self.temp_dir / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint=">=2.0.0",
            description="Test description",
            output="/tmp/output",
            jobs=[
                JobTaskConfig(
                    task="test_task",
                    folders=[Path("/tmp/data")],
                ),
            ],
        )

        config.to_yaml(output_path)

        self.assertTrue(output_path.exists())

        with output_path.open("r") as f:
            data = yaml.safe_load(f)

        self.assertEqual(data["version"], "1.0")
        self.assertEqual(data["hazen_version_constraint"], ">=2.0.0")
        self.assertEqual(data["description"], "Test description")
        self.assertEqual(
            data["output"],
            Path("/tmp/output").absolute().as_posix(),
        )

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_handles_none_values(self, mock_exists: Mock) -> None:
        """Verify None values are written to output."""
        mock_exists.return_value = True
        output_path = self.temp_dir / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint=None,
            description="Test",
            output="/tmp/output",
            jobs=[],
            report_docx=None,
            report_template=None,
            defaults=None,
        )

        config.to_yaml(output_path)

        with output_path.open("r") as f:
            data = yaml.safe_load(f)

        # None values should be present in output
        self.assertIsNone(data["hazen_version_constraint"])
        self.assertIsNone(data["report_docx"])
        self.assertIsNone(data["report_template"])

        # Report is automatically set based on level.
        # levels=None => levels=("final", "all") => report=True
        self.assertEqual(data["defaults"], {"report": True})

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_resolves_paths_to_absolute_posix(
        self,
        mock_exists: Mock,
    ) -> None:
        """Verify paths are converted to absolute POSIX format."""
        mock_exists.return_value = True
        output_path = self.temp_dir / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint="1.0.0",
            description="Test",
            output=Path("relative/output"),
            jobs=[
                JobTaskConfig(
                    task="test_task",
                    folders=[Path("/absolute/path"), Path("relative/folder")],
                ),
            ],
        )

        config.to_yaml(output_path)

        with output_path.open("r") as f:
            data = yaml.safe_load(f)

        # All paths should be absolute POSIX format
        self.assertTrue(Path(data["output"]).is_absolute())
        for folder in data["jobs"][0]["folders"]:
            self.assertTrue(Path(folder).is_absolute())

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_includes_job_overrides(self, mock_exists: Mock) -> None:
        """Verify job overrides are included in output."""
        mock_exists.return_value = True
        output_path = self.temp_dir / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint=">=1.0.0",
            description="Test",
            output="/tmp/output",
            jobs=[
                JobTaskConfig(
                    task="test_task",
                    folders=[Path("/tmp/data")],
                    overrides={"param": "value", "count": 42},
                ),
            ],
        )

        config.to_yaml(output_path)

        with output_path.open("r") as f:
            data = yaml.safe_load(f)

        self.assertEqual(data["jobs"][0]["overrides"]["param"], "value")
        self.assertEqual(data["jobs"][0]["overrides"]["count"], 42)

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_creates_parent_directories(
        self,
        mock_exists: Mock,
    ) -> None:
        """Verify parent directories are created for output path."""
        mock_exists.return_value = True
        nested_path = self.temp_dir / "nested" / "dir" / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint=">=1.0.0",
            description="Test",
            output="/tmp/output",
            jobs=[],
        )

        config.to_yaml(nested_path)

        self.assertTrue(nested_path.parent.exists())
        self.assertTrue(nested_path.exists())

    @patch.dict(TASK_REGISTRY, {"test_task": Mock()}, clear=False)
    @patch("pathlib.Path.exists")
    def test_to_yaml_resolves_path_overrides(self, mock_exists: Mock) -> None:
        """Verify Path values in overrides are resolved to POSIX strings."""
        mock_exists.return_value = True
        output_path = self.temp_dir / "output.yaml"

        config = BatchConfig(
            version="1.0",
            hazen_version_constraint="1.0.0",
            description="Test",
            output="/tmp/output",
            jobs=[
                JobTaskConfig(
                    task="test_task",
                    folders=[Path("/tmp/data")],
                    overrides={
                        "output_path": Path("/custom/output"),
                        "template": Path("relative/template.docx"),
                        "count": 42,
                    },
                ),
            ],
        )

        config.to_yaml(output_path)

        with output_path.open("r") as f:
            data = yaml.safe_load(f)

        # Path overrides should be converted to absolute POSIX strings
        overrides = data["jobs"][0]["overrides"]
        self.assertTrue(Path(overrides["output_path"]).is_absolute())
        self.assertTrue(Path(overrides["template"]).is_absolute())
        # Non-path values should remain unchanged
        self.assertEqual(overrides["count"], 42)


if __name__ == "__main__":
    unittest.main()



"""Tests for the discover.py module."""

# ruff: noqa: PT009

# Python imports
import unittest
from pathlib import Path

# Module imports
# Local imports
from hazenlib._version import __version__
from hazenlib.discovery import DiscoveredAcquisition, generate_batch_config
from hazenlib.orchestration import BatchConfig, JobTaskConfig
from tests import TEST_DATA_DIR


class TestGenerateBatchConfig(unittest.TestCase):
    """Test the generated_batch_config function."""

    PATH: Path = TEST_DATA_DIR / "batch_guess"

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the class."""
        if not cls.PATH.exists():
            msg = f"Required path does not exist: {cls.PATH}"
            raise FileNotFoundError(msg)

        ########
        # Jobs #
        ########

        jobs = [
            # ARC ALL #
            JobTaskConfig(
                task="acr_all",
                folders=[
                    cls.PATH / "ACR_T1_2",
                    cls.PATH / "ACR_T2_1",
                    cls.PATH / "ACR_SAG_LOCALISER_2",
                ],
            ),
            # SNR BODY #
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_BODY_AX_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_BODY_AX_ACQ2"},
            ),
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_BODY_COR_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_BODY_COR_ACQ2"},
            ),
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_BODY_SAG_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_BODY_SAG_ACQ2"},
            ),
            # SNR BREAST #
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_BREAST_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_BREAST_ACQ2"},
            ),
            # SNR HEADNECK #
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_HEADNECK_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_HEADNECK_ACQ2"},
            ),
            # SNR KNEE #
            JobTaskConfig(
                task="snr",
                folders=[cls.PATH / "SNR_KNEE_ACQ1"],
                overrides={"subtract": cls.PATH / "SNR_KNEE_ACQ2"},
            ),
        ]

        for job in jobs:
            for folder in job.folders:
                if not folder.exists():
                    msg = f"Required path does not exist: {folder}"
                    raise FileNotFoundError(msg)
            if (
                job.overrides
                and (sub_dir := job.overrides.get("subtract"))
                and not sub_dir.exists()
            ):
                msg = f"Required path does not exist: {sub_dir}"
                raise FileNotFoundError(msg)

        ###################################
        # True Batch Configuration Object #
        ###################################

        cls.batch_config = BatchConfig(
            version=BatchConfig._CURRENT_BATCHCONFIG_VERSION,  # noqa: SLF001
            hazen_version_constraint=f">={__version__}",
            description=(
                "Batch configuration file automatically generated"
                f" from {cls.PATH.as_posix()}"
            ),
            jobs=jobs,
            output=cls.PATH.parent / "hazen_output.json",
            report_docx=cls.PATH.parent / "hazen_report.docx",
            report_template=None,
            levels=("final", "all"),
            defaults={},
            _file=cls.PATH.parent / "hazen_batch_config.yml",
        )

        #####################################
        # Loaded Batch Configuration Object #
        #####################################

        cls.generated_batch_config = generate_batch_config(cls.PATH)

    def test_generate_batch_config(self) -> None:
        """Test the generate batch config."""
        self.assertTrue(generate_batch_config(self.PATH))

    def test_version(self) -> None:
        """Test Batch Config version."""
        self.assertEqual(
            self.batch_config.version,
            self.generated_batch_config.version,
        )

    def test_hazen_version_constraint(self) -> None:
        """Test Batch Config hazen version contraint."""
        self.assertEqual(
            self.batch_config.hazen_version_constraint,
            self.generated_batch_config.hazen_version_constraint,
        )

    def test_description(self) -> None:
        """Test Batch Config description."""
        self.assertEqual(
            self.batch_config.description,
            self.generated_batch_config.description,
        )

    def test_jobs(self) -> None:
        """Test Batch Config jobs."""

        def normalise_job(job: JobTaskConfig) -> tuple[tuple]:
            """Convert job into a hashable tuple."""
            return (
                job.task,
                sorted(job.folders),
                tuple(sorted(job.overrides.items())) if job.overrides else (),
            )

        for task in {j.task for j in self.batch_config.jobs}:
            expected = [
                normalise_job(j)
                for j in self.batch_config.jobs
                if j.task == task
            ]
            actual = [
                normalise_job(j)
                for j in self.generated_batch_config.jobs
                if j.task == task
            ]
            self.assertCountEqual(
                expected,
                actual,
                f"Jobs not equal for task: {task}",
            )

    def test_output(self) -> None:
        """Test Batch Config output."""
        self.assertEqual(
            self.batch_config.output,
            self.generated_batch_config.output,
        )

    def test_report_docx(self) -> None:
        """Test Batch Config report_docx."""
        self.assertEqual(
            self.batch_config.report_docx,
            self.generated_batch_config.report_docx,
        )

    def test_report_template(self) -> None:
        """Test Batch Config report_template."""
        self.assertEqual(
            self.batch_config.report_template,
            self.generated_batch_config.report_template,
        )

    def test_levels(self) -> None:
        """Test Batch Config levels."""
        self.assertEqual(
            self.batch_config.levels,
            self.generated_batch_config.levels,
        )

    def test_defaults(self) -> None:
        """Test Batch Config defaults."""
        self.assertEqual(
            self.batch_config.defaults,
            self.generated_batch_config.defaults,
        )


class TestIsLikelySNR(unittest.TestCase):
    """Tests for difflib-based fuzzy SNR matching."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the class."""
        cls.is_snr = staticmethod(
            DiscoveredAcquisition._is_likely_snr,  # noqa: SLF001
        )

    def test_fuzzy_matches(self) -> None:
        """Test similarity threshold catches transpositions."""
        # These all have SequenceMatcher ratio >= 0.66 with "snr"
        fuzzy_typos = ["srn", "nsr", "smr"]  # smr is close on keyboard
        for typo in fuzzy_typos:
            with self.subTest(typo=typo):
                result = self.is_snr(typo)
                self.assertTrue(result, f"Should detect {typo} as SNR-like")

        not_snr = [
            "diffusion",
            "phase",
            "sag",
            "axial",
            "coronal",
            "soon",
        ]
        for seq in not_snr:
            with self.subTest(typo=seq):
                result = self.is_snr(seq)
                self.assertFalse(
                    result,
                    f"Should not detect {seq} as SNR-like",
                )


if __name__ == "__main__":
    unittest.main()
